from typing import Callable, Tuple, Type

from vlutils.base import Registry

from mcquic.baseClass import ValueTuner

__all__ = [
    "ModuleRegistry",
    "ValueTunerRegistry",
    "LrSchedulerRegistry",
    "OptimizerRegistry",
    "LossRegistry"
]

class ModuleRegistry(Registry[Type["torch.nn.Module"]]):
    pass

class ValueTunerRegistry(Registry[Type[ValueTuner]]):
    pass

class LrSchedulerRegistry(Registry[Type["torch.optim.lr_scheduler._LRScheduler"]]):
    pass

class OptimizerRegistry(Registry[Type["torch.optim.optimizer.Optimizer"]]):
    pass

class LossRegistry(Registry[Type["torch.nn.Module"]]):
    pass



import pathlib
from shutil import copy2
from typing import Tuple
import os

import torch
from torch import nn
import torch.distributed as dist
from vlutils.config import summary

from mcquic import Config, Consts
from mcquic.modules.compressor import BaseCompressor, Compressor
from mcquic.datasets import getTrainLoader, getValLoader
from mcquic.utils.registry import OptimizerRegistry, LrSchedulerRegistry, LossRegistry
import mcquic.train.lrSchedulers as _
import mcquic.loss as _

from .utils import getSaver, initializeBaseConfigs
from .trainer import getTrainer


def registerForTrain():
    try:
        import apex
        OptimizerRegistry.register("Lamb")(apex.optimizers.FusedLAMB)
    except:
        raise ImportError("`import apex` failed. Apex not installed.")
    OptimizerRegistry.register("Adam")(torch.optim.Adam)

    LrSchedulerRegistry.register("ReduceLROnPlateau")(torch.optim.lr_scheduler.ReduceLROnPlateau)
    LrSchedulerRegistry.register("Exponential")(torch.optim.lr_scheduler.ExponentialLR)
    LrSchedulerRegistry.register("MultiStep")(torch.optim.lr_scheduler.MultiStepLR)
    LrSchedulerRegistry.register("OneCycle")(torch.optim.lr_scheduler.OneCycleLR)


def modelFn(modelParams, lossTarget) -> Tuple[BaseCompressor, nn.Module]:
    compressor = Compressor(**modelParams)
    criterion = LossRegistry.get(lossTarget)()

    return compressor, criterion


def ddpSpawnTraining(rank: int, worldSize: int, port: str, config: Config, saveDir: str, resume: pathlib.Path, loggingLevel: int):
    registerForTrain()


    # load ckpt before create trainer, in case it moved to other place.
    if resume is not None:
        if rank == 0:
            tmpFile = copy2(resume, os.path.join(Consts.TempDir, "resume.ckpt"), follow_symlinks=False)
        else:
            tmpFile = os.path.join(Consts.TempDir, "resume.ckpt")
    else:
        tmpFile = None


    saver = getSaver(saveDir, saveName="saved.ckpt", loggerName=Consts.Name, loggingLevel=loggingLevel, config=config.serialize(), reserve=False, disable=rank != 0)

    saver.info("Here is the whole config during this run: \r\n%s", summary(config.serialize()))

    saver.debug("Creating the world...")

    initializeBaseConfigs(port, rank, worldSize, logger=saver)
    saver.debug("Base configs initialized.")

    dist.barrier()

    optimizerFn = OptimizerRegistry.get(config.Train.Optim.Key, logger=saver)
    schdrFn = LrSchedulerRegistry.get(config.Train.Schdr.Key, logger=saver)

    trainer = getTrainer(rank, config, lambda: modelFn(config.Model.Params, config.Train.Target), optimizerFn, schdrFn, saver)

    if tmpFile is not None:
        saver.info("Found ckpt to resume at %s", resume)
        trainer.restoreStates(tmpFile)

    trainLoader, trainSampler = getTrainLoader(rank, worldSize, config.Train.TrainSet, config.Train.BatchSize, logger=saver)
    valLoader = getValLoader(config.Train.ValSet, disable=rank != 0, logger=saver)
    saver.debug("Train and validation datasets mounted.")

    trainer.train(trainLoader, trainSampler, valLoader)

    saver.debug(summary(config.serialize()))
    saver.info("Bye.")



from typing import List, Tuple, Any

import torch
from torch.distributions import Categorical, kl_divergence
from torchvision.transforms.functional import resize, center_crop, convert_image_dtype
from torchvision.models import inception_v3
from torch.utils.data import DataLoader, Dataset
from vlutils.metrics.meter import Handler

from .metrics import MsSSIM as M, PSNR as P
from .utils import Decibel


class MsSSIM(Handler):
    def __init__(self, format: str = r"%.2f dB"):
        super().__init__(format=format)
        self._msSSIM = M(sizeAverage=False)
        self._formatter = Decibel(1.0)

    def to(self, device):
        self._msSSIM.to(device)
        return super().to(device)

    def handle(self, *, images: torch.ByteTensor, restored: torch.ByteTensor, **_) -> List[float]:
        # [N]
        results: torch.Tensor = self._formatter(self._msSSIM(images.float(), restored.float()))
        return results.tolist()


class PSNR(Handler):
    def __init__(self, format: str = r"%.2f dB"):
        super().__init__(format=format)
        self._psnr = P(sizeAverage=False)

    def to(self, device):
        self._psnr.to(device)
        return super().to(device)

    def handle(self, *, images: torch.ByteTensor, restored: torch.ByteTensor, **_) -> List[float]:
        # [N]
        results: torch.Tensor = self._psnr(images.float(), restored.float())
        return results.tolist()


class BPP(Handler):
    def __init__(self, format: str = r"%.4f"):
        super().__init__(format=format)

    @staticmethod
    def bitLength(byteArray: List[bytes]):
        return sum(len(bi) * 8 for bi in byteArray)

    def handle(self, *, images: torch.ByteTensor, binaries: List[List[bytes]], **_) -> List[float]:
        # binaries: List of binary, len = n, len(binaries[0]) = level
        bits = [self.bitLength(bis) for bis in binaries]
        pixels = images.shape[-2] * images.shape[-1]
        bpps = [bit / pixels for bit in bits]
        return bpps


class Visualization(Handler):
    def __init__(self):
        super().__init__()
        self._temp = None

    def reset(self):
        self._temp = None

    def __call__(self, *args: Any, **kwds: Any):
        self._temp = self.handle(*args, **kwds)

    def handle(self, *, restored: torch.ByteTensor, **_) -> torch.Tensor:
        return restored.detach()

    @property
    def ShowInSummary(self) -> bool:
        return False

    @property
    def Result(self):
        return self._temp

    def __str__(self) -> str:
        return "In Tensorboard."


class ImageCollector(Handler):
    def __init__(self):
        super().__init__()
        self._allImages = list()

    def reset(self):
        self._allImages = list()

    def __call__(self, *args: Any, **kwds: Any):
        self._allImages.extend(self.handle(*args, **kwds))

    def handle(self, *, restored: torch.ByteTensor, stem: List[str], **_) -> List[Tuple[torch.Tensor, str]]:
        return [x for x in zip(restored.detach().cpu(), stem)]

    @property
    def ShowInSummary(self) -> bool:
        return False

    @property
    def Result(self):
        return self._allImages


class IdealBPP(Handler):
    def __init__(self, m: int, k: List[int], format: str = r"%.4f"):
        super().__init__(format)

        self._k = k
        self._m = m
        self.accumulated: List[torch.Tensor] = list(torch.zeros((m, k)) for k in self._k)
        self.totalPixels = torch.zeros(())
        self.totalCodes: List[torch.Tensor] = list(torch.zeros((self._m)) for _ in self._k)

    def reset(self):
        self.length = 0
        self.accumulated = list(torch.zeros((self._m, k)) for k in self._k)
        self.totalPixels = torch.zeros(())
        self.totalCodes = list(torch.zeros((self._m)) for _ in self._k)

    def __call__(self, *args: Any, **kwds: Any):
        results, pixels, codes = self.handle(*args, **kwds)

        # Only give stats of whole dataset
        self.length += 1

        for lv, unqiueCounts in enumerate(results):
            self.accumulated[lv] += unqiueCounts
        self.totalPixels += pixels
        for lv, codeCount in enumerate(codes):
            self.totalCodes[lv] += codeCount

    def handle(self, *, codes: List[torch.Tensor], images: torch.ByteTensor, **_) -> Tuple[List[torch.Tensor], int, List[torch.Tensor]]:
        """[summary]

        Args:
            codes (List[torch.Tensor]): len = level, each code has shape [n, m, h, w]
            images (torch.ByteTensor): [n, c, h, w]

        Returns:
            List[torch.Tensor]: Bincount of codes, len = level, each item has shape [m, k]
            int: Total pixels (n * h * w) of images.
            List[torch.Tensor]]: Total code amount, len = level, each item has shape [m].
        """
        allCounts: List[torch.Tensor] = list()
        codesNum: List[torch.Tensor] = list()
        # [n, m, h, w]
        for code, k in zip(codes, self._k):
            groupCounts = list()
            groupCodeNum = list()
            for m in range(self._m):
                # [n, h, w] -> [k]
                count = torch.bincount(code[:, m].flatten(), minlength=k).cpu()
                groupCounts.append(count)
                groupCodeNum.append(code[:, m].numel())
            # [m, k]
            groupCounts = torch.stack(groupCounts, 0)
            allCounts.append(groupCounts)

            # [m]
            groupCodeNum = torch.tensor(groupCodeNum)
            codesNum.append(groupCodeNum)

        n, _, h, w = images.shape

        # lv * [each image's unique counts, only first group]
        return allCounts, n * h * w, codesNum

    @property
    def Result(self) -> float:
        totalBits = 0.
        for codeUsage, codeCount in zip(self.accumulated, self.totalCodes):
            # [m, k]
            prob = codeUsage / codeUsage.sum(-1, keepdim=True)
            estimateEntropy = prob.log2()
            estimateEntropy[estimateEntropy == float("-inf")] = 0
            # [m]
            estimateEntropy = -(prob * estimateEntropy).sum(-1)
            # [m] * [m] then sum
            totalBits += float((estimateEntropy * codeCount).sum())
        # percentage of usage of all codes
        return totalBits / float(self.totalPixels)

    def __str__(self) -> str:
        return self._format % self.Result


class InceptionScore(Handler):
    def __init__(self):
        super().__init__()
        self._allImages = list()
        self._inceptionModel = inception_v3(pretrained=True, transform_input=False).eval()
        self._rank = "cpu"

    def to(self, device: Any) -> "Handler":
        self._inceptionModel.to(device)
        self._rank = device
        return super().to(device)

    def reset(self):
        self._allImages = list()

    def __call__(self, *args: Any, **kwds: Any):
        self._allImages.extend(self.handle(*args, **kwds))

    def handle(self, *, restored: torch.ByteTensor, **_) -> List[torch.Tensor]:
        return [x for x in restored.detach().cpu()]

    @property
    def Length(self) -> int:
        return len(self._allImages)

    @property
    def Result(self):
        def get_pred(x):
            return self._inceptionModel(x)

        # Get predictions
        preds = list()

        class _dataset(Dataset):
            def __init__(self, orig):
                self.orig = orig

            def __getitem__(self, index):
                return center_crop(resize((convert_image_dtype(self.orig[index], torch.float) - 0.5) * 2, 299), 299)

            def __len__(self):
                return len(self.orig)

        dataLoader = DataLoader(_dataset(self._allImages), 8, True)

        for i, batch in enumerate(dataLoader, 0):
            batch = batch.to(self._rank)
            preds.append(get_pred(batch))

        preds = torch.cat(preds).cpu()

        # Now compute the mean kl-div
        split_scores = []

        for k in range(1):
            part = preds[k * (len(preds) // 1): (k+1) * (len(preds) // 1), ...]
            py = part.mean(0)
            scores = []
            for i in range(part.shape[0]):
                pyx = part[i, :]
                p = Categorical(logits=pyx)
                q = Categorical(logits=py)
                scores.append(kl_divergence(p, q))
            split_scores.append(torch.stack(scores).mean().exp())

        return [torch.stack(split_scores).mean(), torch.stack(split_scores).std()]


class LPips(Handler):
    def __init__(self):
        super().__init__()
        self._metric = LPIPS(net_type='alex', version='0.1')

    def to(self, device: Any) -> "Handler":
        self._metric.to(device)
        return super().to(device)

    def handle(self, images: torch.ByteTensor, restored: torch.ByteTensor, **_) -> List[float]:
        result = self._metric((convert_image_dtype(images, torch.float) - 0.5) * 2, (convert_image_dtype(restored, torch.float) - 0.5) * 2)
        return [result.item()]



from dataclasses import dataclass
import math
from typing import Any, Dict
import json

from marshmallow import Schema, fields, post_load, RAISE


class GeneralSchema(Schema):
    class Meta:
        unknown = RAISE
    key = fields.Str(required=True, description="A unique key used to retrieve in registry. For example, given `Lamb` for optimizers, it will check `OptimRegistry` and find the optimizer `apex.optim.FusedLAMB`.")
    params = fields.Dict(keys=fields.Str(), values=fields.Raw(), required=True, description="Corresponding funcation call parameters. So the whole call is `registry.get(key)(**params)`.", type=["string", "number", "boolean"])

    @post_load
    def _(self, data, **kwargs):
        return General(**data)

class GPUSchema(Schema):
    class Meta:
        unknown = RAISE
    gpus = fields.Int(required=True, description="Number of gpus for training. This affects the `world size` of PyTorch DDP.", exclusiveMinimum=0)
    vRam = fields.Int(required=True, description="Minimum VRam required for each gpu. Set it to `-1` to use all gpus.")
    wantsMore = fields.Bool(required=True, description="Set to `true` to use all visible gpus and all VRams and ignore `gpus` and `vRam`.")

    @post_load
    def _(self, data, **kwargs):
        return GPU(**data)

class TrainSchema(Schema):
    class Meta:
        unknown = RAISE
    batchSize = fields.Int(required=True, description="Batch size for training. NOTE: The actual batch size (whole world) is computed by `batchSize * gpus`.", exclusiveMinimum=0)
    epoch = fields.Int(required=True, description="Total training epochs.", exclusiveMinimum=0)
    valFreq = fields.Int(required=True, description="Run validation after every `valFreq` epochs.", exclusiveMinimum=0)
    trainSet = fields.Str(required=True, description="A dir path to load `lmdb` dataset. You need to convert your images before you give this path by calling `mcquic dataset ...`.")
    valSet = fields.Str(required=True, description="A dir path to load image files for validation.")
    saveDir = fields.Str(required=True, description="A dir path to save model checkpoints, TensorBoard messages and logs.")
    target = fields.Str(required=True, description="Training target. Now is one of `[PSNR, MsSSIM]`.", enum=["PSNR", "MsSSIM"])
    optim = fields.Nested(GeneralSchema(), required=True, description="Optimizer used for training. As for current we have `Adam` and `Lamb`.")
    schdr = fields.Nested(GeneralSchema(), required=True, description="Learning rate scheduler used for training. As for current we have `ReduceLROnPlateau`, `Exponential`, `MultiStep`, `OneCycle` and all schedulers defined in `mcquic.train.lrSchedulers`.")
    gpu = fields.Nested(GPUSchema(), required=True, description="GPU configs for training.")

    @post_load
    def _(self, data, **kwargs):
        return Train(**data)

class ConfigSchema(Schema):
    class Meta:
        unknown = RAISE
    model = fields.Nested(GeneralSchema(), required=True, description="Compression model to use. Now we only have one model, so `key` is ignored. Avaliable params are `channel`, `m` and `k`.")
    train = fields.Nested(TrainSchema(), required=True, description="Training configs.")

    @post_load
    def _(self, data, **kwargs):
        return Config(**data)

@dataclass
class General:
    key: str
    params: Dict[str, Any]

    @property
    def Key(self) -> str:
        return self.key

    @property
    def Params(self) -> Dict[str, Any]:
        return self.params



@dataclass
class GPU:
    gpus: int
    vRam: int
    wantsMore: bool

    @property
    def GPUs(self) -> int:
        return self.gpus

    @property
    def VRam(self) -> int:
        return self.vRam

    @property
    def WantsMore(self) -> bool:
        return self.wantsMore


@dataclass
class Train:
    batchSize: int
    epoch: int
    valFreq: int
    trainSet: str
    valSet: str
    saveDir: str
    target: str
    optim: General
    schdr: General
    gpu: GPU

    @property
    def BatchSize(self) -> int:
        return self.batchSize

    @property
    def Epoch(self) -> int:
        return self.epoch

    @property
    def ValFreq(self) -> int:
        return self.valFreq

    @property
    def TrainSet(self) -> str:
        return self.trainSet

    @property
    def ValSet(self) -> str:
        return self.valSet

    @property
    def SaveDir(self) -> str:
        return self.saveDir

    @property
    def Target(self) -> str:
        return self.target

    @property
    def Optim(self) -> General:
        return self.optim

    @property
    def Schdr(self) -> General:
        return self.schdr

    @property
    def GPU(self) -> GPU:
        return self.gpu


@dataclass
class Config:
    model: General
    train: Train

    @property
    def Model(self) -> General:
        return self.model

    @property
    def Train(self) -> Train:
        return self.train

    def scaleByWorldSize(self, worldSize: int):
        batchSize = self.train.BatchSize * worldSize
        exponent = math.log2(batchSize)
        scale = 3 - exponent / 2
        if "lr" in self.Train.Optim.Params:
            self.Train.Optim.Params["lr"] /= (2 ** scale)

    def serialize(self) -> dict:
        return ConfigSchema().dump(self)

    @staticmethod
    def deserialize(data: dict) -> "Config":
        data = { key: value for key, value in data.items() if "$" not in key }
        return ConfigSchema().load(data)



import os
from typing import List, Optional, Union
from pathlib import Path
from distutils.version import StrictVersion
import warnings
import hashlib

from torch import nn
from rich import filesize
from rich.progress import Progress

import mcquic

from .registry import *

StrPath = Union[str, Path]


def nop(*_, **__):
    pass


def totalParameters(model: nn.Module) -> str:
    allParams = sum(p.numel() for p in model.parameters())
    unit, suffix = filesize.pick_unit_and_suffix(allParams, ["", "k", "M", "B"], 1000)
    return f"{(allParams / unit):.4f}{suffix}"

def bppUpperBound(m: int, k: List[int], featureMapScale: List[float]):
    raise NotImplementedError


def versionCheck(versionStr: str):
    version = StrictVersion(versionStr)
    builtInVersion = StrictVersion(mcquic.__version__)

    if builtInVersion < version:
        raise ValueError(f"Version too new. Given {version}, but I'm {builtInVersion} now.")

    major, minor, revision = version.version

    bMajor, bMinor, bRev = builtInVersion.version

    if major != bMajor:
        raise ValueError(f"Major version mismatch. Given {version}, but I'm {builtInVersion} now.")

    if minor != bMinor:
        warnings.warn(f"Minor version mismatch. Given {version}, but I'm {builtInVersion} now.")
    return True


def hashOfFile(path: StrPath, progress: Optional[Progress] = None):
    sha256 = hashlib.sha256()

    fileSize = os.path.getsize(path)

    if progress is not None:
        task = progress.add_task(f"[ Hash ]", total=fileSize, progress="0.00%", suffix="")

    now = 0

    with open(path, 'rb') as fp:
        while True:
            # Reading is buffered, so we can read smaller chunks.
            chunk = fp.read(65536)
            if not chunk:
                break
            sha256.update(chunk)
            now += 65536
            if progress is not None:
                progress.update(task, advance=65536, progress=f"{now / fileSize * 100 :.2f}%")

    if progress is not None:
        progress.remove_task(task)

    hashResult = sha256.hexdigest()
    return hashResult



import random
import torch
from torch import nn
from torchvision import transforms as T
from torchvision.transforms import functional as F
from torch.distributions import Categorical


__all__ = [
    "RandomGamma",
    "DeTransform",
    "RandomHorizontalFlip",
    "RandomVerticalFlip",
    "RandomAutocontrast",
    "Masking"
]


class BatchRandomGamma(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        randomChoice = torch.randint(0, 4, (len(x), ))
        x[randomChoice == 0] = srgbToLinear(x[randomChoice == 0])
        x[randomChoice == 1] = linearToSrgb(x[randomChoice == 1])
        x[randomChoice == 2] = randomGamma(x[randomChoice == 2], torch.rand((x[randomChoice == 2].shape[0], ), device=x.device) * 1.95 + 0.05)
        return x

def srgbToLinear(x: torch.Tensor):
    return torch.where(x < 0.0031308, 12.92 * x, (1.055 * torch.pow(torch.abs(x), 1 / 2.4) - 0.055))

def linearToSrgb(x: torch.Tensor):
    return torch.where(x < 0.04045, x / 12.92, torch.pow(torch.abs(x + 0.055) / 1.055, 2.4))

def randomGamma(x: torch.Tensor, randomGammas: torch.Tensor):
    x = torch.pow(x.clamp_(0), randomGammas)
    return x.clamp_(0.0, 1.0)

def identity(x: torch.Tensor):
    return x

class RandomGamma(nn.Module):
    _fns = [
        srgbToLinear,
        linearToSrgb,
        lambda x: randomGamma(x, torch.rand((), device=x.device) * 1.95 + 0.05),
        identity
    ]
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        choice = random.randint(0, 3)
        return self._fns[choice](x)


class DeTransform(nn.Module):
    _eps = 1e-3
    _maxVal = 255
    def __init__(self, minValue: float = -1.0, maxValue: float = 1.0):
        super().__init__()
        self._min = float(minValue)
        self._max = float(maxValue)

    def forward(self, x):
        x = (x - self._min) / (self._max - self._min)
        # [0, 1] to [0, 255]
        return (x * (self._maxVal + 1.0 - self._eps)).clamp(0.0, 255.0).byte()


# https://github.com/pratogab/batch-transforms
class RandomHorizontalFlip(nn.Module):
    """Applies the :class:`~torchvision.transforms.RandomHorizontalFlip` transform to a batch of images.
    Args:
        p (float): probability of an image being flipped.
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor of size (N, C, H, W) to be flipped.
        Returns:
            Tensor: Randomly flipped Tensor.
        """
        flipped = torch.rand(tensor.shape[0]) < self.p
        tensor[flipped] = F.hflip(tensor[flipped])
        return tensor

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomVerticalFlip(nn.Module):
    """Applies the :class:`~torchvision.transforms.RandomVerticalFlip` transform to a batch of images.
    Args:
        p (float): probability of an image being flipped.
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor of size (N, C, H, W) to be flipped.
        Returns:
            Tensor: Randomly flipped Tensor.
        """
        flipped = torch.rand(tensor.shape[0]) < self.p
        tensor[flipped] = F.vflip(tensor[flipped])
        return tensor

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomAutocontrast(torch.nn.Module):
    """Autocontrast the pixels of the given image randomly with a given probability.
    If the image is torch Tensor, it is expected
    to have [..., 1 or 3, H, W] shape, where ... means an arbitrary number of leading dimensions.
    If img is PIL Image, it is expected to be in mode "L" or "RGB".
    Args:
        p (float): probability of the image being autocontrasted. Default value is 0.5
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, img):
        """
        Args:
            img (PIL Image or Tensor): Image to be autocontrasted.
        Returns:
            PIL Image or Tensor: Randomly autocontrasted image.
        """
        picked = torch.rand(img.shape[0]) < self.p
        img[picked] = F.autocontrast(img[picked])
        return img

    def __repr__(self):
        return self.__class__.__name__ + f"(p={self.p})"


class Masking(nn.Module):
    def __init__(self):
        super().__init__()
        self._categorical = Categorical(torch.Tensor([0.85, 0.15]))

    def forward(self, images: torch.Tensor):
        n, _, h, w = images.shape
        zeros = torch.zeros_like(images)
        # [n, 1, h, w]
        mask = self._categorical.sample((n, 1, h, w)).byte().to(images.device)
        return (mask == 0) * images + (mask == 1) * zeros



# Copyright 2020 InterDigital Communications, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# https://github.com/InterDigitalInc/CompressAI/blob/master/compressai/layers/layers.py


# Copyright (c) 2020 Patrick Esser and Robin Rombach and Björn Ommer

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE./
# https://github.com/CompVis/taming-transformers/blob/master/taming/modules/diffusionmodules/model.py


# NOTE: Slightly modified based on [CompVis/taming-transformers] for res-blocks, GroupNorm removed.

from math import sqrt
from typing import Union

import torch
from torch import nn

from mcquic.utils import ModuleRegistry
from .gdn import GenDivNorm, InvGenDivNorm
from .convs import MaskedConv2d, conv1x1, conv3x3, pixelShuffle3x3, pixelShuffle1x1


__all__ = [
    "ResidualBlockWithStride",
    "ResidualBlockShuffle",
    "ResidualBlock",
    "ResidualBlockMasked",
    "AttentionBlock",
    "NonLocalBlock"
]


class _residulBlock(nn.Module):
    def __init__(self, act1: nn.Module, conv1: nn.Conv2d, act2: nn.Module, conv2: nn.Conv2d, skip: Union[nn.Module, None]):
        super().__init__()
        self._branch = nn.Sequential(
            act1, conv1, act2, conv2
        )
        self._skip = skip

    def forward(self, x):
        identity = x
        out = self._branch(x)

        if self._skip is not None:
            identity = self._skip(x)

        out += identity
        return out


@ModuleRegistry.register
class ResidualBlockWithStride(_residulBlock):
    """Residual block with stride for down-sampling.

    Default structure:
    ```plain
        +--------------+
        | Input ----╮  |
        | SiLU      |  |
        | Conv3s2   |  |
        | GDN       |  |
        | Conv3s1   |  |
        | + <-------╯  |
        | Output       |
        +--------------+
    ```
    """
    def __init__(self, inChannels: int, outChannels: int, stride: int = 2, groups: int = 1):
        """Usage:
        ```python
            # A block performs 2x down-sampling
            block = ResidualBlockWithStride(128, 128)
        ```
        Args:
            inChannels (int): Channels of input.
            outChannels (int): Channels of output.
            stride (int): stride value (default: 2).
            groups (int): Group convolution (default: 1).
        """
        if stride != 1:
            skip = conv3x3(inChannels, outChannels, stride=stride)
        elif inChannels != outChannels:
            skip = conv1x1(inChannels, outChannels, stride=stride)
        else:
            skip = None
        super().__init__(
            nn.SiLU(),
            conv3x3(inChannels, outChannels, stride=stride),
            GenDivNorm(outChannels, groups=groups),
            conv3x3(outChannels, outChannels),
            skip)

@ModuleRegistry.register
class ResidualBlockShuffle(_residulBlock):
    """Residual block with PixelShuffle for up-sampling.

    Default structure:
    ```plain
        +--------------+
        | Input ----╮  |
        | SiLU      |  |
        | PixShuf3  |  |
        | IGDN      |  |
        | Conv3s1   |  |
        | + <-------╯  |
        | Output       |
        +--------------+
    ```
    """
    def __init__(self, inChannels: int, outChannels: int, upsample: int = 2, groups: int = 1):
        """Usage:
        ```python
            # A block performs 2x up-sampling
            block = ResidualBlockShuffle(128, 128)
        ```
        Args:
            inChannels (int): Channels of input.
            outChannels (int): Channels of output.
            upsample (int): Up-sampling rate (default: 2).
            groups (int): Group convolution (default: 1).
        """
        super().__init__(
            nn.SiLU(),
            pixelShuffle3x3(inChannels, outChannels, upsample),
            InvGenDivNorm(outChannels, groups=groups),
            conv3x3(outChannels, outChannels),
            pixelShuffle3x3(inChannels, outChannels, upsample))


@ModuleRegistry.register
class ResidualBlock(_residulBlock):
    """Basic residual block.

    Default structure:
    ```plain
        +--------------+
        | Input ----╮  |
        | SiLU      |  |
        | Conv3s1   |  |
        | SiLU      |  |
        | Conv3s1   |  |
        | + <-------╯  |
        | Output       |
        +--------------+
    ```
    """
    def __init__(self, inChannels: int, outChannels: int, groups: int = 1):
        """Usage:
        ```python
            # A block with "same" feature map
            block = ResidualBlock(128, 128)
        ```
        Args:
            inChannels (int): Channels of input.
            outChannels (int): Channels of output.
            groups (int): Group convolution (default: 1).
        """
        if inChannels != outChannels:
            skip = conv1x1(inChannels, outChannels)
        else:
            skip = None
        super().__init__(
            nn.SiLU(),
            conv3x3(inChannels, outChannels),
            nn.SiLU(),
            conv3x3(outChannels, outChannels),
            skip)


@ModuleRegistry.register
class ResidualBlockMasked(_residulBlock):
    """A residual block with MaskedConv for causal inference.

    Default structure:
    ```plain
        +--------------+
        | Input ----╮  |
        | MConv5s1A |  |
        | LReLU     |  |
        | MConv5s1B |  |
        | LReLU     |  |
        | + <-------╯  |
        | Output       |
        +--------------+
    ```
    """
    def __init__(self, inChannels, outChannels, maskType: str = "A"):
        """Usage:
        ```python
            # First block
            block = ResidualBlockMasked(128, 128, "A")
            # Subsequent blocks
            block = ResidualBlockMasked(128, 128, "B")
        ```
        Args:
            inChannels (int): Channels of input.
            outChannels (int): Channels of output.
            maskType (str): Mask type of MaskedConv2D (default: "A").
        """
        if inChannels != outChannels:
            skip = MaskedConv2d(inChannels, outChannels, maskType=maskType, bias=False, kernel_size=5, stride=1, padding=5 // 2, padding_mode="zeros")
        else:
            skip = None
        super().__init__(
            nn.ReLU(),
            MaskedConv2d(inChannels, outChannels, maskType=maskType, bias=False, kernel_size=5, stride=1, padding=5 // 2, padding_mode="zeros"),
            nn.ReLU(),
            MaskedConv2d(outChannels, outChannels, maskType="B", bias=False, kernel_size=5, stride=1, padding=5 // 2, padding_mode="zeros"),
            skip)


@ModuleRegistry.register
class AttentionBlock(nn.Module):
    """Self attention block.
    Simplified variant from `"Learned Image Compression with
    Discretized Gaussian Mixture Likelihoods and Attention Modules"
    <https://arxiv.org/abs/2001.01568>`_, by Zhengxue Cheng, Heming Sun, Masaru
    Takeuchi, Jiro Katto.

    Default structure:
    ```plain
        +----------------------+
        | Input ----┬--------╮ |
        | ResBlock  ResBlock | |
        | |         Sigmoid  | |
        | * <------ Mask     | |
        | Masked             | |
        | + <----------------╯ |
        | Output               |
        +----------------------+
    ```
    """
    def __init__(self, channel, groups=1):
        super().__init__()
        self._mainBranch = nn.Sequential(
            ResidualBlock(channel, channel, groups),
            ResidualBlock(channel, channel, groups),
            ResidualBlock(channel, channel, groups)
        )

        self._sideBranch = nn.Sequential(
            ResidualBlock(channel, channel, groups),
            ResidualBlock(channel, channel, groups),
            ResidualBlock(channel, channel, groups),
            conv1x1(channel, channel)
        )

    def forward(self, x):
        identity = x
        a = self._mainBranch(x)
        b = self._sideBranch(x)
        mask = torch.sigmoid(b)
        out = a * mask
        out += identity
        return out


@ModuleRegistry.register
class NonLocalBlock(nn.Module):
    def __init__(self, N, groups=1):
        super().__init__()
        self._c = N // 2
        self._q = conv1x1(N, N // 2, groups=groups)
        self._k = conv1x1(N, N // 2, groups=groups)
        self._v = conv1x1(N, N // 2, groups=groups)
        self._z = conv1x1(N // 2, N, groups=groups)

    def forward(self, x: torch.Tensor):
        n, c, h, w = x.shape
        # x = self._position(x)
        hw = h*w
        scale = sqrt(hw)
        # [n, c/2, h, w]
        q = self._q(x).reshape(n, self._c, hw)
        k = self._k(x).reshape(n, self._c, hw)
        # [n, c/2, h, w] -> [n, hw, c/2]
        v = self._v(x).reshape(n, self._c, hw).permute(0, 2, 1)
        # [n, hw, hw]
        qkLogits = torch.matmul(q.transpose(-1, -2), k) / scale
        randomMask = torch.rand((n, hw, hw), device=qkLogits.device) < 0.1
        qkLogits = qkLogits.masked_fill(randomMask, -1e9)
        weights = torch.softmax(qkLogits, -1)
        # [n, hw, c/2] -> [n, c/2, h, w]
        z = torch.matmul(weights, v).permute(0, 2, 1).reshape(n, self._c, h, w)
        z = self._z(z)
        return x + z



import os
import warnings
import click
import pathlib
import logging

import torch
from vlutils.logger import configLogging
from torchvision.io.image import write_png

import mcquic
from mcquic.config import Config
from mcquic.modules.compressor import Compressor
from mcquic.train.utils import getRichProgress
from mcquic.datasets import getValLoader
from mcquic.utils import hashOfFile, versionCheck

from .validator import Validator


def checkArgs(debug: bool, quiet: bool):
    if quiet:
        return logging.CRITICAL
    if debug:
        return logging.DEBUG
    return logging.INFO


def main(debug: bool, quiet: bool, export: pathlib.Path, path: pathlib.Path, images: pathlib.Path, output: pathlib.Path):
    loggingLevel = checkArgs(debug, quiet)

    logger = configLogging(None, "root", loggingLevel)

    checkpoint = torch.load(path, "cuda")

    config = Config.deserialize(checkpoint["config"])

    model = Compressor(**config.Model.Params).cuda().eval()

    if "trainer" in checkpoint:
        modelStateDict = {key[len("module._compressor."):]: value for key, value in checkpoint["trainer"]["_model"].items()}
    else:
        modelStateDict = checkpoint["model"]
        if export is not None:
            warnings.warn("I got an already-converted ckpt.")
        if not "version" in checkpoint:
            raise RuntimeError("You are using a too old version of ckpt, since there is no `version` in it.")
        versionCheck(checkpoint["version"])

    model.load_state_dict(modelStateDict)

    validator = Validator(config, "cuda")

    valLoader = getValLoader(images, False, logger)

    progress = getRichProgress()

    with progress:
        results, summary = validator.validate(None, model, valLoader, progress)
        logger.info(summary)
        _, speedSummary = validator.speed(None, model, progress)
        logger.info(speedSummary)

        if output is not None:
            allImages = results["ImageCollector"]

            total = len(allImages)

            task = progress.add_task(f"[ Save ]", total=total, progress=f"{0:4d}/{total:4d}", suffix="")

            for now, (image, stem) in enumerate(allImages):
                write_png(image, os.path.join(output, f"{stem}.png"))

                progress.update(task, advance=1, progress=f"{(now + 1):4d}/{total:4d}")
            progress.remove_task(task)

    if export is None:
        logger.info(f"Skip saving model.")
        return

    qp = config.Model.Params["m"]
    finalName = export.joinpath(f"qp_{qp}_{config.Train.Target.lower()}.mcquic")

    torch.save({
        "model": model.state_dict(),
        "config": config.serialize(),
        "version": mcquic.__version__
    }, finalName)

    logger.info(f"Saved at `{finalName}`.")
    logger.info("Add hash to file...")

    with progress:
        hashResult = hashOfFile(finalName, progress)

    newName = f"{finalName.stem}_{hashResult[:8]}{finalName.suffix}"

    os.rename(finalName, finalName.parent.joinpath(newName))

    logger.info("Rename file to %s", newName)


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-e", "--export", type=click.Path(exists=False, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, help="Dir to export the final model that is compatible with main program. Model name is generated automatically.")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("images", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("output", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def entryPoint(debug, quiet, export, path, images, output):
    """Validate a trained model from `path` by images from `images` dir, and publish a final state_dict to `output` path.

Args:

    path (str): Saved checkpoint path.

    images (str): Validation images folder.

    output (str): Dir to save all restored images.
    """
    main(debug, quiet, export, path, images, output)



from torch import nn
import torch.nn.functional as F

from mcquic.validate.metrics import MsSSIM as _m
from mcquic.validate.utils import Decibel
from mcquic.utils.registry import LossRegistry


class Distortion(nn.Module):
    def __init__(self, formatter: nn.Module):
        super().__init__()
        self._formatter = formatter

    def formatDistortion(self, loss):
        return self._formatter(loss)


@LossRegistry.register
class MsSSIM(Distortion):
    def __init__(self):
        super().__init__(Decibel(1.0))
        self._ssim = _m(data_range=2.0, sizeAverage=True)

    def forward(self, restored, image, *_):
        return 0.0, self._ssim(restored + 1, image + 1)

@LossRegistry.register
class PSNR(Distortion):
    def __init__(self):
        super().__init__(Decibel(2.0))

    def forward(self, restored, image, *_):
        return 0.0, F.mse_loss(restored, image)



from dataclasses import dataclass
from typing import List, Union

import msgpack
from marshmallow import Schema, fields, post_load, ValidationError
import vlutils.logger

from mcquic.utils import versionCheck


# TODO: goto marshmallow


class BytesField(fields.Field):
    def _validate(self, value):
        if not isinstance(value, bytes):
            raise ValidationError('Invalid input type.')
        if value is None or value == b'':
            raise ValidationError('Invalid value')


class ImageSizeSchema(Schema):
    height = fields.Int()
    width = fields.Int()
    channel = fields.Int()
    @post_load
    def _(self, data, **kwargs):
        return ImageSize(**data)

class CodeSizeSchema(Schema):
    m = fields.Int()
    heights = fields.List(fields.Int())
    widths = fields.List(fields.Int())
    k = fields.List(fields.Int())
    @post_load
    def _(self, data, **kwargs):
        return CodeSize(**data)

class FileHeaderSchema(Schema):
    qp = fields.Str()
    version = fields.Str()
    codeSize = fields.Nested(CodeSizeSchema())
    imageSize = fields.Nested(ImageSizeSchema())
    @post_load
    def _(self, data, **kwargs):
        return FileHeader(**data)


class FileSchema(Schema):
    fileHeader = fields.Nested(FileHeaderSchema())
    contents = fields.List(BytesField())
    @post_load
    def _(self, data, **kwargs):
        return File(**data)

@dataclass
class ImageSize:
    """Image size specification.

    Args:
        height (int): Image height.
        width (int): Image width.
        channel (int): Image channel.
    """
    height: int
    width: int
    channel: int

    @property
    def Pixels(self) -> int:
        return self.height * self.width

    def __str__(self) -> str:
        return f"[{self.width}×{self.height}, {self.channel}]"


@dataclass
class CodeSize:
    """Latent code specification.
           Code in this paper is of shape: `[[1, m, h, w], [1, m, h, w] ... ]`
                                                            `↑ total length = L`

    Args:
        heights (List[int]): Latent height for each stage.
        widths (List[int]): Latent width for each stage.
        k (List[int]): [k1, k2, ...], codewords amount for each stage.
        m (int): M, multi-codebook amount.
    """
    m: int
    heights: List[int]
    widths: List[int]
    k: List[int]

    def __str__(self) -> str:
        sequence = ", ".join(f"[{w}×{h}, {k}]" for h, w, k in zip(self.heights, self.widths, self.k))
        return f"""
        {self.m} code-groups: {sequence}"""


@dataclass(init=False)
class FileHeader:
    qp: str
    version: str
    codeSize: CodeSize
    imageSize: ImageSize
    def __init__(self, version: str, qp: str, codeSize: CodeSize, imageSize: ImageSize) -> None:
        if versionCheck(version):
            self.qp = qp
            self.version = version
            self.codeSize = codeSize
            self.imageSize = imageSize

    @property
    def QuantizationParameter(self) -> str:
        return str(self.qp)

    @property
    def Version(self) -> str:
        return self.version

    @property
    def CodeSize(self) -> CodeSize:
        return self.codeSize

    @property
    def ImageSize(self) -> ImageSize:
        return self.imageSize

    def __str__(self) -> str:
        return f"""
    Version    : {self.Version}
    QP         : {self.QuantizationParameter}
    Image size : {self.ImageSize}
    Code size  : {self.CodeSize}"""

@dataclass
class File:
    fileHeader: FileHeader
    contents: List[bytes]

    @property
    def FileHeader(self):
        return self.fileHeader

    @property
    def Content(self):
        return self.contents

    def serialize(self) -> bytes:
        thisFile: dict = FileSchema().dump(self)
        return msgpack.packb(thisFile, use_bin_type=True)

    @staticmethod
    def deserialize(data: bytes) -> "File":
        thisFile = msgpack.unpackb(data, use_list=False, raw=False)
        return FileSchema().load(thisFile)

    @property
    def BPP(self) -> float:
        return sum(len(x) for x in self.contents) * 8 / self.FileHeader.ImageSize.Pixels

    def size(self, human: bool = False) -> Union[int, str]:
        """Compute size of compressed binary, in bytes.

        Args:
            human (bool, optional): Whether to give a human-readable string (like `-h` option in POSIX). Defaults to False.

        Returns:
            Union[int, str]: If `human` is True, return integer of total bytes, else return human-readable string.
        """
        size = sum(len(x) for x in self.contents)
        if not human:
            return size
        return vlutils.logger.readableSize(size)

    def __str__(self) -> str:
        return f"""Header: {self.fileHeader}
Size  : {self.size(True)}
BPP   : {self.BPP:.4f}"""

    def __hash__(self) -> int:
        return hash(self.serialize())



import functools
import os
import shutil
from time import sleep
from typing import Callable, Optional, Tuple, Type
import signal
import threading
import gc

import torch
from torch import nn
from torch.utils.data import DataLoader, DistributedSampler
from torch import distributed as dist
from vlutils.base.freqHook import ChainHook
from vlutils.saver import Saver
from vlutils.logger import trackingFunctionCalls
from vlutils.base import Restorable
from vlutils.runtime import relativePath
from vlutils.config import serialize
from rich import filesize

from mcquic.consts import Consts
from mcquic.validate.utils import Decibel, EMATracker
from mcquic.modules.composed import Composed
from mcquic import Config
from mcquic.modules.compressor import BaseCompressor
from mcquic.validate import Validator
from mcquic.utils import totalParameters

from .utils import EpochFrequencyHook, checkHook, getRichProgress

class _baseTrainer(Restorable):
    def __init__(self, config: Config, modelFn: Callable[[], Tuple[BaseCompressor, nn.Module]], optimizer: Type[torch.optim.Optimizer], scheduler: Type[torch.optim.lr_scheduler._LRScheduler], saver: Saver, **_) -> None:
        super().__init__()
        self.saver = saver

        self.rank = dist.get_rank()
        self.saver.debug("<%s> is located at rank `%d`", self.__class__.__name__, self.rank)
        self.worldSize = dist.get_world_size()
        torch.cuda.set_device(self.rank)
        self.config = config

        self._epoch = 0
        self._step = 0

        # Used for self.PrettyStep
        self.lastFormatted = -1
        self.prettyStep = "     "

        self.saver.debug("[%s] Creating model...", self.PrettyStep)
        compressor, criterion = trackingFunctionCalls(modelFn, self.saver)()
        self._model = Composed(compressor.to(self.rank), criterion.to(self.rank), device_ids=[self.rank], output_device=self.rank)
        self.saver.debug("[%s] Model created.", self.PrettyStep)
        self.saver.debug("[%s] Model size: %s", self.PrettyStep, totalParameters(self._model))

        self.saver.debug("[%s] Creating optimizer...", self.PrettyStep)
        optimizer = trackingFunctionCalls(optimizer, self.saver)
        self._optimizer = optimizer(self._model.parameters(), **self.config.Train.Optim.Params)
        self.optimFn = optimizer
        self.saver.debug("[%s] Optimizer created.", self.PrettyStep)

        self.saver.debug("[%s] Creating LR scheduler...", self.PrettyStep)
        scheduler = trackingFunctionCalls(scheduler, self.saver)
        self._scheduler = scheduler(self._optimizer, **self.config.Train.Schdr.Params)
        self.schdrFn = scheduler
        self.saver.debug("[%s] LR scheduler created.", self.PrettyStep)

        self.saver.debug("[%s] <%s> created.", self.PrettyStep, self.__class__.__name__)

    def save(self, path = None):
        self.saver.save(path, trainer=self, config=self.config.serialize())

    @property
    def PrettyStep(self):
        if self._step == self.lastFormatted:
            return self.prettyStep
        else:
            self.prettyStep = self._formatStep(self._step)
            self.lastFormatted = self._step
            return self.prettyStep

    @staticmethod
    def _formatStep(step):
        unit, suffix = filesize.pick_unit_and_suffix(step, ["", "k", "M"], 1000)
        if unit < 10:
            return f"{(step // unit):5d}"
        else:
            truncated = step / unit
            if truncated < 10:
                return f"{truncated:4.6f}"[:4] + suffix
            elif truncated < 100:
                return f"{truncated:4.6f}"[:4] + suffix
            else:
                return f"{truncated:11.6f}"[:4] + suffix

    def restoreStates(self, path: str):
        self.saver.debug("[%s] Restored state dict from `%s`", self.PrettyStep, path)

        self.saver.load(path, torch.device(f"cuda:{self.rank}"), logger=self.saver, trainer=self)

        self.saver.debug("[%s] Restore network parameters finished.", self.PrettyStep)

        self.resetOptimizer()
        for group in self._optimizer.param_groups:
            group.setdefault('initial_lr', group['lr'])

        self.resetScheduler(self._scheduler.last_epoch)

    def resetOptimizer(self):
        del self._optimizer
        self._optimizer = self.optimFn(self._model.parameters(), **self.config.Train.Optim.Params)

        self.saver.debug("[%s] Optimizer reset.", self.PrettyStep)

    def resetScheduler(self, lastEpoch=-1):
        del self._scheduler
        self._scheduler = self.schdrFn(self._optimizer, last_epoch=lastEpoch, **self.config.Train.Schdr.Params)

        self.saver.debug("[%s] LR scheduler reset.", self.PrettyStep)

    def _beforeRun(self, hook, *args, totalBatches, **kwargs):
        hook(self._step, self._epoch, *args, totalBatches=totalBatches, **kwargs)
        if self._step > 0:
            self.saver.info("[%s] Resume training at %s steps / %d epochs.", self.PrettyStep, self.PrettyStep, self._epoch)
        else:
            self.saver.info("[%s] Start training.", self.PrettyStep)

        self.saver.debug("[%s] Training loop started.", self.PrettyStep)

    def _afterRun(self, hook, *args, **kwArgs):
        hook(self._step, self._epoch, *args, **kwArgs)
        self.saver.debug("[%s] Training loop finished.", self.PrettyStep)

    def _stepStart(self, hook, *args, **kwArgs):
        hook(self._step, self._epoch, *args, **kwArgs)

    def _stepFinish(self, hook, *args, **kwArgs):
        self._step += 1
        hook(self._step, self._epoch, *args, **kwArgs)

    def _epochStart(self, hook, *args, trainSampler, **kwArgs):
        trainSampler.set_epoch(self._epoch)
        hook(self._step, self._epoch, *args, trainSampler, **kwArgs)

        self.saver.debug("[%s] Epoch %4d started.", self.PrettyStep, self._epoch + 1)

        gc.collect()
        gc.collect()

    def _epochFinish(self, hook, *args, trainSet, **kwArgs):
        self._epoch += 1

        self.saver.debug("[%s] Epoch %4d finished.", self.PrettyStep, self._epoch)

        self._scheduler.step()
        self.saver.debug("[%s] Lr is set to %.2e.", self.PrettyStep, self._scheduler.get_last_lr()[0])

        hook(self._step, self._epoch, *args, trainSet=trainSet, **kwArgs)

        if self._epoch % (self.config.Train.ValFreq) == 0:
            self.refresh()

    @torch.inference_mode()
    def refresh(self, *_, **__):
        self.saver.debug("[%s] Start refresh at epoch %4d.", self.PrettyStep, self._epoch)

        reAssignProportion = self._model.refresh(self.rank)

        self.saver.debug("[%s] %.2f%% of codebook is re-assigned.", self.PrettyStep, reAssignProportion * 100)

        self.saver.debug("[%s] End refresh at epoch %4d.", self.PrettyStep, self._epoch)
        return reAssignProportion

    def train(self, trainLoader: DataLoader, trainSampler: DistributedSampler, *_, beforeRunHook: Optional[Callable] = None, afterRunHook: Optional[Callable] = None, epochStartHook: Optional[Callable] = None, epochFinishHook: Optional[Callable] = None, stepStartHook: Optional[Callable] = None, stepFinishHook: Optional[Callable] = None, **__):
        beforeRunHook = checkHook(beforeRunHook, "BeforeRunHook", self.saver)
        afterRunHook = checkHook(afterRunHook, "AfterRunHook", self.saver)
        stepStartHook = checkHook(stepStartHook, "StepStartHook", self.saver)
        stepFinishHook = checkHook(stepFinishHook, "StepFinishHook", self.saver)
        epochStartHook = checkHook(epochStartHook, "EpochStartHook", self.saver)
        epochFinishHook = checkHook(epochFinishHook, "EpochFinishHook", self.saver)

        self._beforeRun(beforeRunHook, totalBatches=len(trainLoader))

        for _ in range(self._epoch, self.config.Train.Epoch):
            self._epochStart(epochStartHook, trainSampler=trainSampler)
            for images in trainLoader:
                images = images.to(self.rank, non_blocking=True)

                self._stepStart(stepStartHook)

                self._optimizer.zero_grad()

                xHat, (rate, distortion), codes, logits = self._model(images)
                (rate + distortion).backward()

                self._optimizer.step()

                self._stepFinish(stepFinishHook, rate=rate, distortion=distortion)
            self._epochFinish(epochFinishHook, images=images, restored=xHat, codes=codes, logits=logits, trainSet=trainLoader.dataset)
        self._afterRun(afterRunHook)


class MainTrainer(_baseTrainer):
    def __init__(self, config: Config, modelFn: Callable[[], Tuple[BaseCompressor, nn.Module]], optimizer: Type[torch.optim.Optimizer], scheduler: Type[torch.optim.lr_scheduler._LRScheduler], saver: Saver) -> None:
        if dist.get_rank() != 0:
            raise AttributeError("A sub-process should not to be a <MainTrainer>, use <PalTrainer> instead.")

        self.rank = dist.get_rank()
        self.config = config
        self.saver = saver

        self.progress = getRichProgress().__enter__()
        self.trainingBar = self.progress.add_task("", start=False, progress="[----/----]", suffix=Consts.CDot * 10)
        self.epochBar = self.progress.add_task("[----/----]", start=False, progress="", suffix=Consts.CDot * 10)

        self.validator = Validator(self.config, self.rank)

        self.diffTracker = EMATracker((), 0.99).to(self.rank)

        # Call function at every X epoches.
        self.epochFinishCalls = EpochFrequencyHook(
            (1, self.log),
            logger=self.saver
        )
        self.epochStartCalls = EpochFrequencyHook(
            (self.config.Train.ValFreq, self.validate),
            logger=self.saver
        )

        self.bestRate = 1e10
        self.bestDistortion = -1

        super().__init__(config, modelFn, optimizer, scheduler, saver)

        signal.signal(signal.SIGTERM, self._terminatedHandler)

    def _kill(self):
        sleep(Consts.TimeOut)
        self.saver.critical("[%s] Timeout exceeds, killed.", self.PrettyStep)
        signal.raise_signal(signal.SIGKILL)

    # Handle SIGTERM when main process is terminated.
    # Save necessary info.
    def _terminatedHandler(self, signum, frame):
        killer = threading.Thread(target=self._kill, daemon=True)
        killer.start()
        self.saver.critical("[%s] Main process was interrupted, try to save necessary info.", self.PrettyStep)
        self.saver.critical("[%s] This post-process will be killed after %d secs if stuck.", self.PrettyStep, Consts.TimeOut)
        self.progress.__exit__(None, None, None)
        self.save(os.path.join(self.saver.SaveDir, "last.ckpt"))
        self.saver.critical("[%s] Find the last checkpoint at `%s`", self.PrettyStep, relativePath(os.path.join(self.saver.SaveDir, "last.ckpt")))
        self.summary()
        self.saver.critical("[%s] QUIT.", self.PrettyStep)
        # reset to default SIGTERM handler
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.raise_signal(signal.SIGTERM)

    def summary(self):
        if abs(self.bestRate / self.bestDistortion) > 1e4:
            self.saver.info("[%s] Total epoches: %d, total steps: %s, best rate/distortion: N/A.", self.PrettyStep, self._epoch, self.PrettyStep)
        else:
            self.saver.info("[%s] Total epoches: %d, total steps: %s, best rate/distortion: %.4f / %.2fdB.", self.PrettyStep, self._epoch, self.PrettyStep, self.bestRate, self.bestDistortion)
        self.saver.info("[%s] Test this model by `python -m mcquic.validate --path %s`.", self.PrettyStep, relativePath(os.path.join(self.saver.SaveDir, "[ONE_OF_A].ckpt")))

    def train(self, trainLoader: DataLoader, trainSampler: DistributedSampler, valLoader: DataLoader, *_, beforeRunHook: Optional[Callable] = None, afterRunHook: Optional[Callable] = None, epochStartHook: Optional[Callable] = None, epochFinishHook: Optional[Callable] = None, stepStartHook: Optional[Callable] = None, stepFinishHook: Optional[Callable] = None, **__):
        return super().train(trainLoader, trainSampler,
            beforeRunHook=beforeRunHook,
            afterRunHook=afterRunHook,
            epochStartHook=ChainHook(
                functools.partial(self.epochStartCalls, valLoader=valLoader),
                epochStartHook),
            epochFinishHook=ChainHook(
                self.epochFinishCalls,
                epochFinishHook),
            stepStartHook=stepStartHook,
            stepFinishHook=ChainHook(self._stepFinishHook, stepFinishHook))

    def _beforeRun(self, hook, *args, totalBatches, **kwargs):
        self.progress.update(self.trainingBar, total=totalBatches)
        self.progress.update(self.epochBar, total=self.config.Train.Epoch * totalBatches, completed=self._step, description=f"[{self._epoch + 1:4d}/{self.config.Train.Epoch:4d}]")
        self.progress.start_task(self.trainingBar)
        self.progress.start_task(self.epochBar)

        super()._beforeRun(hook, *args, totalBatches=totalBatches, **kwargs)
        self.saver.info("[%s] See you at `%s`", self.PrettyStep, self.saver.TensorboardURL)

    def _afterRun(self, hook, *args, **kwArgs):
        self.progress.__exit__(None, None, None)
        super()._afterRun(hook, *args, **kwArgs)
        self.summary()

    def _stepFinishHook(self, *_, rate, distortion, **__):
        distortionDB = self._model.formatDistortion(distortion)
        moment = self.diffTracker(distortionDB)

        task = self.progress.get_task(self.trainingBar)
        self.progress.update(self.trainingBar, advance=1, progress=f"[{task.completed + 1:4d}/{task.total:4d}]", suffix=f"D = [b green]{moment:2.2f}[/]dB")
        self.progress.update(self.epochBar, advance=1)

        if self._step % 100 != 0:
            return
        self.saver.add_scalar(f"Stat/{self.config.Train.Target}", moment, global_step=self._step)
        self.saver.add_scalar("Stat/Lr", self._scheduler.get_last_lr()[0], global_step=self._step)

    def _epochStart(self, hook, *args, **kwArgs):
        self.progress.reset(self.trainingBar)
        self.progress.update(self.epochBar, description=f"[{self._epoch + 1:4d}/{self.config.Train.Epoch:4d}]")
        super()._epochStart(hook, *args, **kwArgs)

    def refresh(self, *_, **__):
        reAssignProportion = super().refresh()
        self.saver.add_scalar("Stat/ReAssignProportion", reAssignProportion, global_step=self._step)

    def log(self, *_, images, restored, codes, logits, **__):
        self.saver.add_scalar("Stat/Epoch", self._epoch, self._step)
        # First level, first image, first group
        self.saver.add_histogram("Stat/LogDistance", (-(logits[0][0, 0])).clamp(Consts.Eps).log10(), global_step=self._step)
        freq = self._model.Compressor.Freq
        # [m, ki]
        for lv, (fr, c) in enumerate(zip(freq, codes)):
            self.saver.add_histogram_raw(f"Stat/FreqLv{lv}", min=0, max=len(fr[0]), num=len(fr[0]), sum=fr[0].sum(), sum_squares=(fr[0] ** 2).sum(), bucket_limits=list(range(len(fr[0]))), bucket_counts=fr[0], global_step=self._step)
            self.saver.add_images(f"Train/CodeLv{lv}", self.validator.visualizeIntermediate(c), self._step)
        self.saver.add_images("Train/Raw", self.validator.tensorToImage(images), global_step=self._step)
        self.saver.add_images("Train/Res", self.validator.tensorToImage(restored), global_step=self._step)
        self.saver.add_scalar("Stat/CodeUsage", self._model.Compressor.CodeUsage, global_step=self._step)

    def validate(self, *_, valLoader: DataLoader, **__):
        torch.cuda.empty_cache()

        self.saver.debug("[%s] Start validation at epoch %4d.", self.PrettyStep, self._epoch)

        self._model.eval()
        results, summary = self.validator.validate(self._epoch, self._model.Compressor, valLoader, self.progress)

        self.saver.add_scalar(f"Eval/MsSSIM", results["MsSSIM"], global_step=self._step)
        self.saver.add_scalar(f"Eval/PSNR", results["PSNR"], global_step=self._step)
        self.saver.add_scalar(f"Eval/BPP", results["BPP"], global_step=self._step)
        self.saver.add_images(f"Eval/Visualization", results["Visualization"], global_step=self._step)

        self.save()

        rate, distortion = results["BPP"], results[self.config.Train.Target]

        # TODO: Why d/r continously decrease?
        # if (distortion / rate) > (self.bestDistortion / self.bestRate):
        if distortion > self.bestDistortion:
            self.bestDistortion = distortion
            self.bestRate = rate
            self.progress.update(self.epochBar, suffix=f"H = [b red]{self.bestDistortion:2.2f}[/]dB")
            shutil.copy2(self.saver.SavePath, os.path.join(self.saver.SaveDir, "best.ckpt"))
        self.saver.info("[%s] %s", self.PrettyStep, summary)
        self._model.train()

        self.saver.debug("[%s] End validation at epoch %4d.", self.PrettyStep, self._epoch)


class PalTrainer(_baseTrainer):
    def __init__(self, config: Config, modelFn: Callable[[], Tuple[BaseCompressor, nn.Module]], optimizer: Type[torch.optim.Optimizer], scheduler: Type[torch.optim.lr_scheduler._LRScheduler], saver: Saver) -> None:
        if dist.get_rank() == 0:
            raise AttributeError("You should call <MainTrainer> for main process other than <PalTrainer> to save, log necessary information.")
        super().__init__(config, modelFn, optimizer, scheduler, saver)

    def train(self, trainLoader: DataLoader, trainSampler: DistributedSampler, *_, beforeRunHook: Optional[Callable] = None, afterRunHook: Optional[Callable] = None, epochStartHook: Optional[Callable] = None, epochFinishHook: Optional[Callable] = None, stepStartHook: Optional[Callable] = None, stepFinishHook: Optional[Callable] = None, **__):
        return super().train(trainLoader, trainSampler, beforeRunHook=beforeRunHook, afterRunHook=afterRunHook, epochStartHook=epochStartHook, epochFinishHook=epochFinishHook, stepStartHook=stepStartHook, stepFinishHook=stepFinishHook)


def getTrainer(rank: int, config: Config, model: Callable[[], Tuple[BaseCompressor, nn.Module]], optimizer: Type[torch.optim.Optimizer], scheduler: Type[torch.optim.lr_scheduler._LRScheduler], saver: Saver) -> _baseTrainer:
    if rank == 0:
        return MainTrainer(config, model, optimizer, scheduler, saver)
    return PalTrainer(config, model, optimizer, scheduler, saver)



from typing import Union
import torch
from torch import nn

from mcquic import Consts

__all__ = [
    "NonNegativeParametrizer",
    "LogExpMinusOne",
    "logExpMinusOne"
]


class _lowerBound(torch.autograd.Function):
    """Autograd function for the `LowerBound` operator."""

    @staticmethod
    def forward(ctx, input_, bound):
        ctx.save_for_backward(input_, bound)
        return torch.max(input_, bound)

    @staticmethod
    def backward(ctx, grad_output):
        input_, bound = ctx.saved_tensors
        pass_through_if = (input_ >= bound) | (grad_output < 0)
        return pass_through_if.type(grad_output.dtype) * grad_output, None

class LowerBound(nn.Module):
    """Lower bound operator, computes `torch.max(x, bound)` with a custom
    gradient.
    The derivative is replaced by the identity function when `x` is moved
    towards the `bound`, otherwise the gradient is kept to zero.
    """

    def __init__(self, bound: float):
        """Lower bound operator.

        Args:
            bound (float): The lower bound.
        """
        super().__init__()
        self.register_buffer("bound", torch.Tensor([float(bound)]))

    @torch.jit.unused
    def lower_bound(self, x):
        return _lowerBound.apply(x, self.bound)

    def forward(self, x):
        if torch.jit.is_scripting():
            return torch.max(x, self.bound)
        return self.lower_bound(x)


class NonNegativeParametrizer(nn.Module):
    """
    Non negative reparametrization.
    Used for stability during training.
    """
    def __init__(self, minimum: float = 0.0, eps: float = Consts.Eps):
        """Non negative reparametrization.

        Args:
            minimum (float, optional): The lower bound. Defaults to 0.
            reparam_offset (float, optional): Eps for stable training. Defaults to 2**-18.
        """
        super().__init__()

        minimum = float(minimum)
        eps = float(eps)

        self.register_buffer("eps", torch.Tensor([eps ** 2]))
        bound = (minimum + eps ** 2) ** 0.5
        self.lowerBound = LowerBound(bound)

    def init(self, x):
        return torch.sqrt(torch.max(x + self.eps, self.eps))

    def forward(self, x):
        out = self.lowerBound(x)
        out = out ** 2 - self.eps
        return out


class _logExpMinusOne(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, bound):
        ctx.save_for_backward(x, bound)
        return (x.exp() - 1 + Consts.Eps).log()

    @staticmethod
    def backward(ctx, grad_output):
        x, bound = ctx.saved_tensors
        passThroughIf = x > bound
        remaining = ~passThroughIf
        return passThroughIf * grad_output + remaining * grad_output * x.exp() / (x.exp() - 1 + Consts.Eps), None

class LogExpMinusOne(nn.Module):
    def __init__(self, eps: Union[torch.Tensor, float] = Consts.Eps) -> None:
        super().__init__()
        eps = torch.tensor(eps, dtype=torch.float)
        self.register_buffer("_bound", ((1 + eps) / eps).log())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return _logExpMinusOne.apply(x, self._bound)

def logExpMinusOne(x: torch.Tensor, eps: Union[torch.Tensor, float] = Consts.Eps) -> torch.Tensor:
    eps = torch.tensor(eps, dtype=torch.float, device=x.device)
    return _logExpMinusOne.apply(x, ((1 + eps) / eps).log())


def gumbelArgmaxRandomPerturb(logits: torch.Tensor, perturbRate: float = 0.0, tau: float = 1, dim: int = -1) -> torch.Tensor:
    r"""
    Samples from the Gumbel-Argmax distribution and optionally perturb (modify sample to another result).

    Args:
      logits: `[..., num_features]` unnormalized log probabilities
      perturbRate: how many samples needs perturbation
      tau: non-negative scalar temperature
      dim (int): A dimension along which softmax will be computed. Default: -1.

    Returns:
      Sampled tensor of same shape as `logits` from the Gumbel-Softmax distribution.
      If ``hard=True``, the returned samples will be one-hot, otherwise they will
      be probability distributions that sum to 1 across `dim`.

    .. note::
      This function is here for legacy reasons, may be removed from nn.Functional in the future.

    .. note::
      The main trick for `hard` is to do  `y_hard - y_soft.detach() + y_soft`

      It achieves two things:
      - makes the output value exactly one-hot
      (since we add then subtract y_soft value)
      - makes the gradient equal to y_soft gradient
      (since we strip all other gradients)

    Examples::
        >>> logits = torch.randn(20, 32)
        >>> # Sample soft categorical using reparametrization trick:
        >>> F.gumbel_softmax(logits, tau=1, hard=False)
        >>> # Sample hard categorical using "Straight-through" trick:
        >>> F.gumbel_softmax(logits, tau=1, hard=True)
    """
    gumbels = (
        -torch.empty_like(logits, memory_format=torch.legacy_contiguous_format).exponential_().log()
    )  # ~Gumbel(0,1)
    gumbels = (logits + gumbels) / tau  # ~Gumbel(logits,tau)
    y_soft = gumbels.softmax(dim)

    # Straight through.
    index = y_soft.max(dim, keepdim=True)[1]

    perturbCount = round(index.numel() * perturbRate)
    if perturbCount > 0:
        newIdx = torch.randint(logits.shape[dim], size=(perturbCount, ), device=index.device, dtype=index.dtype)

        # inplace opertaion
        continguous = index.view(-1)
        permIdx = torch.randperm(len(continguous))
        continguous[permIdx[:perturbCount]] = newIdx

    y_hard = torch.zeros_like(logits, memory_format=torch.legacy_contiguous_format).scatter_(dim, index, 1.0)
    ret = y_hard - y_soft.detach() + y_soft
    return ret

def oneHot(x: torch.LongTensor, numClasses: int, dim: int = -1, dtype = torch.float):
    return torch.zeros((*x.shape, numClasses), dtype=dtype).scatter_(dim, x, 1)



import logging
import pathlib
import random

import torch
import torch.multiprocessing as mp
import click
import yaml
from vlutils.runtime import queryGPU

from mcquic.config import Config

def checkArgs(debug, quiet, resume: pathlib.Path, configPath: pathlib.Path):
    if resume is None and configPath is None:
        raise ValueError("Both `resume` and `config` are not given.")
    if quiet:
        return logging.CRITICAL
    if debug:
        return logging.DEBUG
    return logging.INFO

def main(debug: bool, quiet: bool, resume: pathlib.Path, configPath: pathlib.Path) -> int:
    assert False, "You need to run `mcquic train` with Python optimized-mode. Try re-run me with `python -O -m mcquic.train ...`"
    loggingLevel = checkArgs(debug, quiet, resume, configPath)

    from .ddp import ddpSpawnTraining, registerForTrain

    logging.getLogger().setLevel(loggingLevel)

    if configPath is not None:
        config = Config.deserialize(yaml.full_load(configPath.read_text()))
    elif resume is not None:
        ckpt = torch.load(resume, "cpu")
        config = Config.deserialize(ckpt["config"])
    else:
        raise ValueError("Both `--resume` and `config` are None.")

    gpus = queryGPU(needGPUs=config.Train.GPU.GPUs, wantsMore=config.Train.GPU.WantsMore, needVRamEachGPU=(config.Train.GPU.VRam + 256) if config.Train.GPU.VRam > 0 else -1, writeOSEnv=True)
    worldSize = len(gpus)
    config.scaleByWorldSize(worldSize)

    masterPort = str(random.randint(10001, 65535))

    registerForTrain()

    # `daemon` is True --- Way to handle SIGINT globally.
    # Give up handling SIGINT by yourself... PyTorch hacks it.
    mp.spawn(ddpSpawnTraining, (worldSize, masterPort, config, config.Train.SaveDir, resume, loggingLevel), worldSize, daemon=True)
    return 0


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-r", "--resume", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1, help="`.ckpt` file path to resume training.")
@click.argument('config', type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def entryPoint(debug, quiet: bool, resume: pathlib.Path, config: pathlib.Path):
    main(debug, quiet, resume, config)




from collections import Counter
from functools import wraps
import math
import warnings
import weakref

import torch
from torch.optim.optimizer import Optimizer

from mcquic.utils import LrSchedulerRegistry


@LrSchedulerRegistry.register
class MultiStepLRWithWarmUp(torch.optim.lr_scheduler._LRScheduler):
    """Decays the learning rate of each parameter group by gamma once the
    number of epoch reaches one of the milestones. Notice that such decay can
    happen simultaneously with other changes to the learning rate from outside
    this scheduler. When last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        milestones (list): List of epoch indices. Must be increasing.
        gamma (float): Multiplicative factor of learning rate decay.
            Default: 0.1.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.

    Example:
        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.05     if epoch < 30
        >>> # lr = 0.005    if 30 <= epoch < 80
        >>> # lr = 0.0005   if epoch >= 80
        >>> scheduler = MultiStepLR(optimizer, milestones=[30,80], gamma=0.1)
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """

    def __init__(self, optimizer, milestones, gamma=0.1, last_epoch=-1, verbose=False):
        self.milestones = Counter(milestones)
        self.gamma = gamma
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        firstMilestone = min(self.milestones)

        if self.last_epoch <= firstMilestone:
            return [base_lr * float(self.last_epoch + 1) / firstMilestone for base_lr in self.base_lrs]

        if self.last_epoch not in self.milestones:
            return [group['lr'] for group in self.optimizer.param_groups]
        return [group['lr'] * self.gamma
                for group in self.optimizer.param_groups]


@LrSchedulerRegistry.register
class CyclicLR(torch.optim.lr_scheduler._LRScheduler):
    r"""Sets the learning rate of each parameter group according to
    cyclical learning rate policy (CLR). The policy cycles the learning
    rate between two boundaries with a constant frequency, as detailed in
    the paper `Cyclical Learning Rates for Training Neural Networks`_.
    The distance between the two boundaries can be scaled on a per-iteration
    or per-cycle basis.

    Cyclical learning rate policy changes the learning rate after every batch.
    `step` should be called after a batch has been used for training.

    This class has three built-in policies, as put forth in the paper:

    * "triangular": A basic triangular cycle without amplitude scaling.
    * "triangular2": A basic triangular cycle that scales initial amplitude by half each cycle.
    * "exp_range": A cycle that scales initial amplitude by :math:`\text{gamma}^{\text{cycle iterations}}`
      at each cycle iteration.

    This implementation was adapted from the github repo: `bckenstler/CLR`_

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        base_lr (float or list): Initial learning rate which is the
            lower boundary in the cycle for each parameter group.
        max_lr (float or list): Upper learning rate boundaries in the cycle
            for each parameter group. Functionally,
            it defines the cycle amplitude (max_lr - base_lr).
            The lr at any cycle is the sum of base_lr
            and some scaling of the amplitude; therefore
            max_lr may not actually be reached depending on
            scaling function.
        step_size_up (int): Number of training iterations in the
            increasing half of a cycle. Default: 2000
        step_size_down (int): Number of training iterations in the
            decreasing half of a cycle. If step_size_down is None,
            it is set to step_size_up. Default: None
        mode (str): One of {triangular, triangular2, exp_range}.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
            Default: 'triangular'
        gamma (float): Constant in 'exp_range' scaling function:
            gamma**(cycle iterations)
            Default: 1.0
        scale_fn (function): Custom scaling policy defined by a single
            argument lambda function, where
            0 <= scale_fn(x) <= 1 for all x >= 0.
            If specified, then 'mode' is ignored.
            Default: None
        scale_mode (str): {'cycle', 'iterations'}.
            Defines whether scale_fn is evaluated on
            cycle number or cycle iterations (training
            iterations since start of cycle).
            Default: 'cycle'
        cycle_momentum (bool): If ``True``, momentum is cycled inversely
            to learning rate between 'base_momentum' and 'max_momentum'.
            Default: True
        base_momentum (float or list): Lower momentum boundaries in the cycle
            for each parameter group. Note that momentum is cycled inversely
            to learning rate; at the peak of a cycle, momentum is
            'base_momentum' and learning rate is 'max_lr'.
            Default: 0.8
        max_momentum (float or list): Upper momentum boundaries in the cycle
            for each parameter group. Functionally,
            it defines the cycle amplitude (max_momentum - base_momentum).
            The momentum at any cycle is the difference of max_momentum
            and some scaling of the amplitude; therefore
            base_momentum may not actually be reached depending on
            scaling function. Note that momentum is cycled inversely
            to learning rate; at the start of a cycle, momentum is 'max_momentum'
            and learning rate is 'base_lr'
            Default: 0.9
        last_epoch (int): The index of the last batch. This parameter is used when
            resuming a training job. Since `step()` should be invoked after each
            batch instead of after each epoch, this number represents the total
            number of *batches* computed, not the total number of epochs computed.
            When last_epoch=-1, the schedule is started from the beginning.
            Default: -1
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.

    Example:
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> scheduler = torch.optim.lr_scheduler.CyclicLR(optimizer, base_lr=0.01, max_lr=0.1)
        >>> data_loader = torch.utils.data.DataLoader(...)
        >>> for epoch in range(10):
        >>>     for batch in data_loader:
        >>>         train_batch(...)
        >>>         scheduler.step()


    .. _Cyclical Learning Rates for Training Neural Networks: https://arxiv.org/abs/1506.01186
    .. _bckenstler/CLR: https://github.com/bckenstler/CLR
    """

    def __init__(self,
                 optimizer,
                 base_lr,
                 max_lr,
                 step_size_up=2000,
                 step_size_down=None,
                 mode='triangular',
                 gamma=1.,
                 scale_fn=None,
                 scale_mode='cycle',
                 cycle_momentum=True,
                 base_momentum=0.8,
                 max_momentum=0.9,
                 last_epoch=-1,
                 verbose=False):

        # Attach optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        base_lrs = self._format_param('base_lr', optimizer, base_lr)
        if last_epoch == -1:
            for lr, group in zip(base_lrs, optimizer.param_groups):
                group['lr'] = lr

        self.max_lrs = self._format_param('max_lr', optimizer, max_lr)

        step_size_up = float(step_size_up)
        step_size_down = float(step_size_down) if step_size_down is not None else step_size_up
        self.total_size = step_size_up + step_size_down
        self.step_ratio = step_size_up / self.total_size

        if mode not in ['triangular', 'triangular2', 'exp_range'] \
                and scale_fn is None:
            raise ValueError('mode is invalid and scale_fn is None')

        self.mode = mode
        self.gamma = gamma

        if scale_fn is None:
            if self.mode == 'triangular':
                self.scale_fn = self._triangular_scale_fn
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = self._triangular2_scale_fn
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = self._exp_range_scale_fn
                self.scale_mode = 'iterations'
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode

        self.cycle_momentum = cycle_momentum
        if self.cycle_momentum:
            if 'momentum' not in self.optimizer.defaults and 'betas' not in self.optimizer.defaults:
                raise ValueError('optimizer must support momentum with `cycle_momentum` option enabled')
            self.use_beta1 = 'betas' in self.optimizer.defaults
            max_momentums = self._format_param('max_momentum', optimizer, max_momentum)
            base_momentums = self._format_param('base_momentum', optimizer, base_momentum)
            if last_epoch == -1:
                for m_momentum, b_momentum, group in zip(max_momentums, base_momentums, optimizer.param_groups):
                    if self.use_beta1:
                        _, beta2 = group['betas']
                        group['betas'] = (m_momentum, beta2)
                    else:
                        group['momentum'] = m_momentum
                    group['max_momentum'] = m_momentum
                    group['base_momentum'] = b_momentum
            # self.base_momentums = base_momentums
            # self.max_momentums = max_momentums
        super(CyclicLR, self).__init__(optimizer, last_epoch, verbose)
        self.base_lrs = base_lrs

    def _format_param(self, name, optimizer, param):
        """Return correctly formatted lr/momentum for each param group."""
        if isinstance(param, (list, tuple)):
            if len(param) != len(optimizer.param_groups):
                raise ValueError("expected {} values for {}, got {}".format(
                    len(optimizer.param_groups), name, len(param)))
            return param
        else:
            return [param] * len(optimizer.param_groups)

    def _triangular_scale_fn(self, x):
        return 1.

    def _triangular2_scale_fn(self, x):
        return 1 / (2. ** (x - 1))

    def _exp_range_scale_fn(self, x):
        return self.gamma**(x)

    def get_lr(self):
        """Calculates the learning rate at batch index. This function treats
        `self.last_epoch` as the last batch index.

        If `self.cycle_momentum` is ``True``, this function has a side effect of
        updating the optimizer's momentum.
        """

        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        cycle = math.floor(1 + self.last_epoch / self.total_size)
        x = 1. + self.last_epoch / self.total_size - cycle
        if x <= self.step_ratio:
            scale_factor = x / self.step_ratio
        else:
            scale_factor = (x - 1) / (self.step_ratio - 1)

        lrs = []
        for base_lr, max_lr in zip(self.base_lrs, self.max_lrs):
            base_height = (max_lr - base_lr) * scale_factor
            if self.scale_mode == 'cycle':
                lr = base_lr + base_height * self.scale_fn(cycle)
            else:
                lr = base_lr + base_height * self.scale_fn(self.last_epoch)
            lrs.append(lr)

        if self.cycle_momentum:
            for group in self.optimizer.param_groups:
                base_momentum, max_momentum = group["base_momentum"], group["max_momentum"]
                base_height = (max_momentum - base_momentum) * scale_factor
                if self.scale_mode == 'cycle':
                    momentum = max_momentum - base_height * self.scale_fn(cycle)
                else:
                    momentum = max_momentum - base_height * self.scale_fn(self.last_epoch)
                if self.use_beta1:
                    _, beta2 = group['betas']
                    group['betas'] = (momentum, beta2)
                else:
                    group['momentum'] = momentum
        return lrs



@LrSchedulerRegistry.register
class CosineAnnealingWarmupRestarts(torch.optim.lr_scheduler._LRScheduler):
    """
        optimizer (Optimizer): Wrapped optimizer.
        first_cycle_steps (int): First cycle step size.
        cycle_mult(float): Cycle steps magnification. Default: -1.
        max_lr(float): First cycle's max learning rate. Default: 0.1.
        min_lr(float): Min learning rate. Default: 0.001.
        warmup_steps(int): Linear warmup step size. Default: 0.
        gamma(float): Decrease rate of max learning rate by cycle. Default: 1.
        last_epoch (int): The index of last epoch. Default: -1.
    """

    def __init__(self,
                 optimizer : torch.optim.Optimizer,
                 first_cycle_steps : int,
                 cycle_mult : float = 1.,
                 lrScaleRatio: float = 0.01,
                 warmup_steps : int = 0,
                 gamma : float = 1.,
                 last_epoch : int = -1,
                 cycle_momentum: bool = False,
                 max_momentum: float = 0.9,
                 base_momentum: float = 0.8,
                 verbose: bool = False
        ):
        assert warmup_steps < first_cycle_steps

        self.optimizer = optimizer

        self.first_cycle_steps = first_cycle_steps # first cycle step size
        self.cycle_mult = cycle_mult # cycle steps magnification
        self.max_lrs = []
        for param_group in optimizer.param_groups:
            self.max_lrs.append(param_group['lr'])
        self.min_lrs = [lr * lrScaleRatio for lr in self.max_lrs] # min learning rate
        self.min_base_lrs = [lr * lrScaleRatio for lr in self.max_lrs] # min learning rate
        self.warmup_steps = warmup_steps # warmup step size
        self.gamma = gamma # decrease rate of max learning rate by cycle

        self.cur_cycle_steps = first_cycle_steps # first cycle step size
        self.cycle = 0 # cycle count
        self.step_in_cycle = last_epoch # step size of the current cycle

        # self.cycle_momentum = cycle_momentum
        # if self.cycle_momentum:
        #     if 'momentum' not in self.optimizer.defaults and 'betas' not in self.optimizer.defaults:
        #         raise ValueError('optimizer must support momentum with `cycle_momentum` option enabled')
        #     self.use_beta1 = 'betas' in self.optimizer.defaults
        #     max_momentums = self._format_param('max_momentum', optimizer, max_momentum)
        #     base_momentums = self._format_param('base_momentum', optimizer, base_momentum)
        #     if last_epoch == -1:
        #         for m_momentum, b_momentum, group in zip(max_momentums, base_momentums, optimizer.param_groups):
        #             if self.use_beta1:
        #                 _, beta2 = group['betas']
        #                 group['betas'] = (m_momentum, beta2)
        #             else:
        #                 group['momentum'] = m_momentum
        #             group['max_momentum'] = m_momentum
        #             group['base_momentum'] = b_momentum

        # Attach optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        # Initialize epoch and base learning rates
        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified "
                                   "in param_groups[{}] when resuming an optimizer".format(i))
        self.base_lrs = [group['initial_lr'] for group in optimizer.param_groups]
        self.last_epoch = last_epoch

        # Following https://github.com/pytorch/pytorch/issues/20124
        # We would like to ensure that `lr_scheduler.step()` is called after
        # `optimizer.step()`
        def with_counter(method):
            if getattr(method, '_with_counter', False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # Keep a weak reference to the optimizer instance to prevent
            # cyclic references.
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned function here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper

        self.optimizer.step = with_counter(self.optimizer.step)
        self.optimizer._step_count = 0
        self._step_count = 0
        self.verbose = verbose

        self.step(None if last_epoch == -1 else last_epoch)

    def _format_param(self, name, optimizer, param):
        """Return correctly formatted lr/momentum for each param group."""
        if isinstance(param, (list, tuple)):
            if len(param) != len(optimizer.param_groups):
                raise ValueError("expected {} values for {}, got {}".format(
                    len(optimizer.param_groups), name, len(param)))
            return param
        else:
            return [param] * len(optimizer.param_groups)

    def get_lr(self):
        if self.step_in_cycle == -1:
            return self.min_lrs
        elif self.step_in_cycle < self.warmup_steps:
            # if self.cycle_momentum:
            #     for group in self.optimizer.param_groups:
            #         base_momentum, max_momentum = group["base_momentum"], group["max_momentum"]
            #         base_height = (max_momentum - base_momentum) * scale_factor
            #         if self.scale_mode == 'cycle':
            #             momentum = max_momentum - base_height * self.scale_fn(cycle)
            #         else:
            #             momentum = max_momentum - base_height * self.scale_fn(self.last_epoch)
            #         if self.use_beta1:
            #             _, beta2 = group['betas']
            #             group['betas'] = (momentum, beta2)
            #         else:
            #             group['momentum'] = momentum

            return [(max_lr - min_lr) * self.step_in_cycle / self.warmup_steps + min_lr for min_lr, max_lr in zip(self.min_lrs, self.max_lrs)]
        else:
            return [min_lr + (max_lr - min_lr) \
                    * (1 + math.cos(math.pi * (self.step_in_cycle-self.warmup_steps) \
                                    / (self.cur_cycle_steps - self.warmup_steps))) / 2
                    for min_lr, max_lr in zip(self.min_lrs, self.max_lrs)]

    def step(self, epoch=None):
        if epoch is None:
            epoch = self.last_epoch + 1
            self.step_in_cycle = self.step_in_cycle + 1
            if self.step_in_cycle >= self.cur_cycle_steps:
                self.cycle += 1
                self.step_in_cycle = self.step_in_cycle - self.cur_cycle_steps
                self.cur_cycle_steps = int((self.cur_cycle_steps - self.warmup_steps) * self.cycle_mult) + self.warmup_steps
                self.max_lrs = [lr * (self.gamma ** self.cycle) for lr in self.base_lrs]
                self.min_lrs = [lr * (self.gamma ** self.cycle) for lr in self.min_base_lrs]
        else:
            if epoch >= self.first_cycle_steps:
                if self.cycle_mult == 1.:
                    self.step_in_cycle = epoch % self.first_cycle_steps
                    self.cycle = epoch // self.first_cycle_steps
                else:
                    n = int(math.log((epoch / self.first_cycle_steps * (self.cycle_mult - 1) + 1), self.cycle_mult))
                    self.cycle = n
                    self.step_in_cycle = epoch - int(self.first_cycle_steps * (self.cycle_mult ** n - 1) / (self.cycle_mult - 1))
                    self.cur_cycle_steps = self.first_cycle_steps * self.cycle_mult ** (n)
                self.max_lrs = [lr * (self.gamma ** self.cycle) for lr in self.base_lrs]
                self.min_lrs = [lr * (self.gamma ** self.cycle) for lr in self.min_base_lrs]
            else:
                self.cur_cycle_steps = self.first_cycle_steps
                self.step_in_cycle = epoch

        super().step()



import math

from mcquic.utils import ValueTunerRegistry
from mcquic.baseClass import ValueTuner

@ValueTunerRegistry.register
class CyclicValue(ValueTuner):
    def __init__(self, initValue: float = 2e-2, gamma: float = 1.0, cyclicInterval: int = 400, boostInterval: int = 3, zeroOutRatio: float = 1./3.):
        super().__init__(initValue=initValue)
        self._cyclicInterval = cyclicInterval
        self._boostInterval = boostInterval
        self._zeroOutRatio = zeroOutRatio
        self._gamma = gamma

    def calc(self):
        maxReg = self._initValue * (self._gamma ** self._epoch)
        # phase 1
        if (self._epoch // self._cyclicInterval) % self._boostInterval == 0:
            self._value = maxReg
        # phase 2
        else:
            j = (self._epoch % self._cyclicInterval) / float(self._cyclicInterval)
            down = 2 * maxReg / (self._zeroOutRatio - 1) * j + maxReg
            up = 2 * maxReg / (1 - self._zeroOutRatio) * j + (self._zeroOutRatio + 1) / (self._zeroOutRatio - 1) * maxReg
            self._value = max(0, max(up, down))


@ValueTunerRegistry.register
class ExponentialValue(ValueTuner):
    def __init__(self, initValue: float = 2e-2, gamma: float = 0.9999):
        super().__init__(initValue=initValue)
        self._gamma = gamma

    def calc(self):
        self._value = self._initValue * (self._gamma ** self._epoch)


@ValueTunerRegistry.register
class StepValue(ValueTuner):
    def __init__(self, initValue: float = 2e-2, gamma: float = 0.1, stepInterval: int = 1000):
        super().__init__(initValue=initValue)
        self._gamma = gamma
        self._stepInterval = stepInterval

    def calc(self):
        self._value = self._initValue * (self._gamma ** (self._epoch // self._stepInterval))

@ValueTunerRegistry.register
class CosineValue(ValueTuner):
    def __init__(self, maxValue: float = 1.0, minValue: float = 0.0, stepInterval: int = 1, totalStep: int = 1000, revert: bool = False):
        super().__init__(initValue=maxValue)
        self._minValue = minValue
        self._stepInterval = stepInterval
        self._totalStep = totalStep
        self._revert = revert

    def calc(self):
        # 1 ~ -1
        nowCosine = math.cos(math.pi * self._epoch / self._stepInterval / self._totalStep)
        if self._revert:
            nowCosine = nowCosine
        realValue = (nowCosine + 1) / 2.0 * (self._initValue - self._minValue) + self._minValue
        self._value = realValue

@ValueTunerRegistry.register
class CosineValueWithEnd(ValueTuner):
    def __init__(self, maxValue: float = 1.0, minValue: float = 0.0, stepInterval: int = 1, totalStep: int = 1000, revert: bool = False):
        super().__init__(initValue=maxValue)
        self._minValue = minValue
        self._stepInterval = stepInterval
        self._totalStep = totalStep
        self._revert = revert

    def calc(self):
        if self._epoch / self._stepInterval / self._totalStep >= 1:
            self._value = self._minValue
            return
        # 1 ~ -1
        nowCosine = math.cos(math.pi * self._epoch / self._stepInterval / self._totalStep)
        if self._revert:
            nowCosine = -nowCosine
        realValule = (nowCosine + 1) / 2.0 * (self._initValue - self._minValue) + self._minValue
        self._value = realValule


@ValueTunerRegistry.register
class JumpValue(ValueTuner):
    def __init__(self, initValue: float = 10.0, gamma: float = 0.9, stepInterval: int = 1000, minValue: float = 0.01):
        super().__init__(initValue=initValue)
        self._gamma = gamma
        self._stepInterval = stepInterval
        self._max = initValue
        self._min = minValue

        self._iteration = int(math.log(self._min / self._max) / math.log(self._gamma))

    def calc(self):
        self._value = self._initValue * (self._gamma ** ((self._epoch // self._stepInterval) % self._iteration))


@ValueTunerRegistry.register
class JumpAlter(ValueTuner):
    def __init__(self, initValue: float = 10.0, gamma: float = 0.9, stepInterval: int = 10, minValue: float = 0.01, milestone: int = 500, valueAfterMilestone: float = 0.01):
        super().__init__(initValue=initValue)
        self._gamma = gamma
        self._stepInterval = stepInterval
        self._max = initValue
        self._min = minValue
        self._milestone = milestone
        self._valueAfterMilestone = valueAfterMilestone

        self._iteration = int(math.log(self._min / self._max) / math.log(self._gamma))

    def calc(self):
        if self._iteration <= self._milestone:
            self._value = self._initValue * (self._gamma ** ((self._epoch // self._stepInterval) % self._iteration))
        else:
            self._value = self._valueAfterMilestone



import abc

from vlutils.base import Restorable


class Serializable(abc.ABC):
    @abc.abstractmethod
    def serialize(self):
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def deserialize(raw):
        raise NotImplementedError


class ValueTuner(Restorable):
    def __init__(self, initValue: float = 2e-2):
        super().__init__()
        self._epoch = 0
        self._initValue = initValue

    def step(self):
        self._epoch += 1
        self.calc()

    def calc(self):
        self._value = self._initValue

    @property
    def Value(self) -> float:
        if not hasattr(self, "_value"):
            self.calc()
        return self._value



import logging
import pathlib
from typing import Optional, Tuple, Union
import warnings

import torch
import torch.hub
from torchvision.io.image import read_image, ImageReadMode, write_png
from torchvision.transforms.functional import convert_image_dtype
from vlutils.logger import configLogging

from mcquic import Config
from mcquic.modules.compressor import BaseCompressor, Compressor
from mcquic.utils import versionCheck
from mcquic.utils.specification import File
from mcquic.utils.vision import DeTransform


MODELS_URL = "https://github.com/xiaosu-zhu/McQuic/releases/download/generic/"

MODELS_HASH = {
    "qp_2_msssim": "8e954998"
}


def checkArgs(debug: bool, quiet: bool):
    if quiet:
        return logging.CRITICAL
    if debug:
        return logging.DEBUG
    return logging.INFO


def main(debug: bool, quiet: bool, qp: int, local: pathlib.Path, disable_gpu: bool, mse: bool, crop: bool, input: pathlib.Path, output: pathlib.Path):
    loggingLevel = checkArgs(debug, quiet)

    if disable_gpu or not torch.cuda.is_available():
        device = torch.device("cpu")
    else:
        device = torch.device("cuda")

    logger = configLogging(None, "root", loggingLevel)

    if input.suffix.lower() in [".png", ".jpg", ".jpeg"]:
        model = loadModel(qp, local, device, mse, logger)

        image = read_image(str(input), ImageReadMode.RGB).to(device)

        target = compressImage(image, model, crop)

        logger.info(target)
        if output is not None:
            if output.is_dir():
                output = output.joinpath(input.stem + ".mcq")
            with open(output, "wb") as fp:
                fp.write(target.serialize())
            logger.info("Saved at %s", output)

    elif input.suffix.lower() == ".mcq":
        with open(input, "rb") as fp:
            binary = fp.read()
            source = File.deserialize(binary)

            model = detectModelFromFile(qp, local, mse, device, logger, source)

            restored = decompressImage(source, model)

            logger.info(source)

            if output is not None:
                if output.is_dir():
                    output = output.joinpath(input.stem + ".png")
                write_png(restored.cpu(), str(output))
    else:
        raise ValueError("Invalid input file.")

def detectModelFromFile(qp, local, mse, device, logger, source):
    localFile = detectLocalFile(source.FileHeader.QuantizationParameter)
    if isinstance(localFile, pathlib.Path):
        model = loadModel(-1, localFile, device, False, logger)
    else:
        parsed = parseQP(source.FileHeader.QuantizationParameter)
        if parsed is not None:
            model = loadModel(parsed[0], None, device, parsed[1], logger)
        else:
            warnings.warn("All qp detections failed. Fallback to use current args or you could try again after checks.")
            model = loadModel(qp, local, device, mse, logger)
    return model


def detectLocalFile(qp: str) -> Union[bool, pathlib.Path]:
    filePath = pathlib.Path(qp)
    if filePath.exists() and filePath.is_file() and "mcquic" in filePath.suffix.lower():
        return filePath
    return False

def parseQP(qp: str) -> Union[None, Tuple[int, bool]]:
    try:
        if not qp.startswith("qp_"):
            return None
        # qp_x_[mse/msssim]
        parsed = qp.split("_")
        return int(parsed[1]), parsed[2] == "mse"
    except:
        return None

def compressImage(image: torch.Tensor, model: BaseCompressor, crop: bool) -> File:
    image = convert_image_dtype(image)

    if crop:
        from mcquic.datasets.transforms import AlignedCrop
        image = AlignedCrop()(image)

    # [c, h, w]
    image = (image - 0.5) * 2

    with model.readyForCoding() as cdfs:
        _, binaries, headers = model.compress(image[None, ...], cdfs)

    # List of each level binary, FileHeader
    return File(headers[0], binaries[0])


def decompressImage(sourceFile: File, model: BaseCompressor) -> torch.Tensor:

    binaries = sourceFile.Content

    with model.readyForCoding() as cdfs:
        # append it to list to make batch-size = 1.
        # [1, c, h, w]
        restored = model.decompress([binaries], cdfs, [sourceFile.FileHeader])

    # [c, h, w]
    return DeTransform()(restored[0])


def loadModel(qp: int, local: Optional[pathlib.Path], device, mse: bool, logger: logging.Logger) -> BaseCompressor:
    # Fallback key
    key = "qp_2_msssim"
    if local is not None:
        warnings.warn(f"By passing `--local`, `-qp` arg will be ignored. Checkpoint from {local} will be loaded. Please ensure you obtain this local model from a trusted source.")
        ckpt = torch.load(local, device)

        logger.info("Use local model.")
    else:
        suffix = "mse" if mse else "msssim"
        key = f"qp_{qp}_{suffix}"
        if key not in MODELS_HASH:
            raise ValueError(f"The provided {key} combination not found in pretrained models.")
        ckpt = torch.hub.load_state_dict_from_url(MODELS_URL + f"qp_{qp}_{suffix}_{MODELS_HASH[key]}.mcquic", map_location=device, check_hash=True)

        logger.info("Use model `-qp %d` targeted `%s`.", qp, suffix)

    if not "version" in ckpt:
        raise RuntimeError("You are using a too old ckpt where `version` not in it.")
    versionCheck(ckpt["version"])

    config = Config.deserialize(ckpt["config"])
    model = Compressor(**config.Model.Params).to(device).eval()
    model.QuantizationParameter = str(local) if local is not None else key
    model.load_state_dict(ckpt["model"])
    logger.info(f"Model loaded, params: {config.Model.Params}.")
    return model



import math
from typing import Callable, Dict, Iterable, List, Tuple, Union

import torch
from torch import nn
import torch.distributed as dist
import torch.nn.functional as F

from mcquic.modules.entropyCoder import EntropyCoder
from mcquic.nn.base import LowerBound
from mcquic.utils.specification import CodeSize
from mcquic import Consts


class BaseQuantizer(nn.Module):
    def __init__(self, m: int, k: List[int]):
        super().__init__()
        self._entropyCoder = EntropyCoder(m, k)
        self._m = m
        self._k = k

    def encode(self, x: torch.Tensor) -> List[torch.Tensor]:
        raise NotImplementedError

    def decode(self, codes: List[torch.Tensor]) -> torch.Tensor:
        raise NotImplementedError

    def readyForCoding(self):
        return self._entropyCoder.readyForCoding()

    def reAssignCodebook(self) -> torch.Tensor:
        raise NotImplementedError

    def syncCodebook(self):
        raise NotImplementedError

    @property
    def Freq(self):
        return self._entropyCoder.Freq

    def compress(self, x: torch.Tensor, cdfs: List[List[List[int]]]) -> Tuple[List[torch.Tensor], List[List[bytes]], List[CodeSize]]:
        codes = self.encode(x)

        # List of binary, len = n, len(binaries[0]) = level
        binaries, codeSize = self._entropyCoder.compress(codes, cdfs)
        return codes, binaries, codeSize

    def _validateCode(self, refCodes: List[torch.Tensor], decompressed: List[torch.Tensor]):
        for code, restored in zip(refCodes, decompressed):
            if torch.any(code != restored):
                raise RuntimeError("Got wrong decompressed result from entropy coder.")

    def decompress(self, binaries: List[List[bytes]], codeSize: List[CodeSize], cdfs: List[List[List[int]]]) -> torch.Tensor:
        decompressed = self._entropyCoder.decompress(binaries, codeSize, cdfs)
        # self._validateCode(codes, decompressed)
        return self.decode(decompressed)


# NOTE: You may notice the quantizer implemented here is different with README.md
#       After some tests, I find some strange behavior if `k` is not placed in the last dim.
#       Generally, although code is neat and output is same as here,
#         training with README's implementation will cause loss become suddenly NAN after a few epoches.
class _multiCodebookQuantization(nn.Module):
    def __init__(self, codebook: nn.Parameter, permutationRate: float = 0.15):
        super().__init__()
        self._m, self._k, self._d = codebook.shape
        self._codebook = codebook
        self._scale = math.sqrt(self._k)
        self._temperature = nn.Parameter(torch.ones((self._m, 1, 1, 1)))
        self._bound = LowerBound(Consts.Eps)
        self._permutationRate = permutationRate

    def reAssignCodebook(self, freq: torch.Tensor)-> torch.Tensor:
        codebook = self._codebook.clone().detach()
        freq = freq.to(self._codebook.device).clone().detach()
        #       [k, d],        [k]
        for m, (codebookGroup, freqGroup) in enumerate(zip(self._codebook, freq)):
            neverAssigned = codebookGroup[freqGroup < 1]
            if len(neverAssigned) > self._k // 2:
                mask = torch.zeros((len(neverAssigned), ), dtype=torch.long, device=self._codebook.device)
                maskIdx = torch.randperm(len(mask))[self._k // 2:]
                mask[maskIdx] = 1
                freqGroup[freqGroup < 1] = mask
                neverAssigned = codebookGroup[freqGroup < 1]
            argIdx = torch.argsort(freqGroup, descending=True)[:(self._k - len(neverAssigned))]
            fullAssigned = codebookGroup[argIdx]
            selectedIdx = torch.randperm(len(fullAssigned))[:len(neverAssigned)]
            codebook.data[m, freqGroup < 1] = fullAssigned[selectedIdx]
        # [m, k] bool
        diff = ((codebook - self._codebook) ** 2).sum(-1) > 1e-6
        proportion = diff.flatten()
        self._codebook.data.copy_(codebook)
        return proportion

    def syncCodebook(self):
        # codebook = self._codebook.clone().detach()
        dist.broadcast(self._codebook, 0)

    def encode(self, x: torch.Tensor):
        # [n, m, h, w, k]
        distance = self._distance(x)
        # [n, m, h, w, k] -> [n, m, h, w]
        code = distance.argmin(-1)
        #      [n, m, h, w]
        return code

    # NOTE: ALREADY CHECKED CONSISTENCY WITH NAIVE IMPL.
    def _distance(self, x: torch.Tensor) -> torch.Tensor:
        n, _, h, w = x.shape
        # [n, m, d, h, w]
        x = x.reshape(n, self._m, self._d, h, w)

        # [n, m, 1, h, w]
        x2 = (x ** 2).sum(2, keepdim=True)

        # [m, k, 1, 1]
        c2 = (self._codebook ** 2).sum(-1, keepdim=True)[..., None]
        # [n, m, d, h, w] * [m, k, d] -sum-> [n, m, k, h, w]
        inter = torch.einsum("nmdhw,mkd->nmkhw", x, self._codebook)
        # [n, m, k, h, w]
        distance = x2 + c2 - 2 * inter
        # IMPORTANT to move k to last dim --- PLEASE SEE NOTE.
        # [n, m, h, w, k]
        return distance.permute(0, 1, 3, 4, 2)

    def _logit(self, x: torch.Tensor) -> torch.Tensor:
        logit = -1 * self._distance(x)
        return logit / self._scale

    def _sample(self, x: torch.Tensor, temperature: float):
        # [n, m, h, w, k] * [m, 1, 1, 1]
        logit = self._logit(x) * self._bound(self._temperature)

        # It causes training unstable
        # leave to future tests.
        # add random mask to pick a different index.
        # [n, m, h, w]
        # needPerm = torch.rand_like(logit[..., 0]) < self._permutationRate * rateScale
        # target will set to zero (one of k) but don't break gradient
        # mask = F.one_hot(torch.randint(self._k, (needPerm.sum(), ), device=logit.device), num_classes=self._k).float() * logit[needPerm]
        # logit[needPerm] -= mask.detach()

        # NOTE: STE: code usage is very low; RelaxedOneHotCat: Doesn't have STE trick
        # So reverse back to F.gumbel_softmax
        # posterior = OneHotCategoricalStraightThrough(logits=logit / temperature)
        # [n, m, k, h, w]
        # sampled = posterior.rsample(())

        sampled = F.gumbel_softmax(logit, temperature, True)

        # It causes training unstable
        # leave to future tests.
        # sampled = gumbelArgmaxRandomPerturb(logit, self._permutationRate * rateScale, temperature)
        return sampled, logit

    def forward(self, x: torch.Tensor):
        sample, logit = self._sample(x, 1.0)
        # [n, m, h, w, 1]
        code = logit.argmax(-1, keepdim=True)
        # [n, m, h, w, k]
        oneHot = torch.zeros_like(logit).scatter_(-1, code, 1)
        # [n, m, h, w, k]
        return sample, code[..., 0], oneHot, logit


class _multiCodebookDeQuantization(nn.Module):
    def __init__(self, codebook: nn.Parameter):
        super().__init__()
        self._m, self._k, self._d = codebook.shape
        self._codebook = codebook
        self.register_buffer("_ix", torch.arange(self._m), persistent=False)

    def decode(self, code: torch.Tensor):
        # codes: [n, m, h, w]
        n, _, h, w = code.shape
        # [n, h, w, m]
        code = code.permute(0, 2, 3, 1)
        # use codes to index codebook (m, k, d) ==> [n, h, w, m, k] -> [n, c, h, w]
        ix = self._ix.expand_as(code)
        # [n, h, w, m, d]
        indexed = self._codebook[ix, code]
        # [n, c, h, w]
        return indexed.reshape(n, h, w, -1).permute(0, 3, 1, 2)

    # NOTE: ALREADY CHECKED CONSISTENCY WITH NAIVE IMPL.
    def forward(self, sample: torch.Tensor):
        n, _, h, w, _ = sample.shape
        # [n, m, h, w, k, 1], [m, 1, 1, k, d] -sum-> [n, m, h, w, d] -> [n, m, d, h, w] -> [n, c, h, w]
        return torch.einsum("nmhwk,mkd->nmhwd", sample, self._codebook).permute(0, 1, 4, 2, 3).reshape(n, -1, h, w)


class _quantizerEncoder(nn.Module):
    """
    Default structure:
    ```plain
        x [H, W]
        | `latentStageEncoder`
        z [H/2 , W/2] -------╮
        | `quantizationHead` | `latentHead`
        q [H/2, W/2]         z [H/2, w/2]
        |                    |
        ├-`subtract` --------╯
        residual for next level
    ```
    """

    def __init__(self, quantizer: _multiCodebookQuantization, dequantizer: _multiCodebookDeQuantization, latentStageEncoder: nn.Module, quantizationHead: nn.Module, latentHead: Union[None, nn.Module]):
        super().__init__()
        self._quantizer = quantizer
        self._dequantizer = dequantizer
        self._latentStageEncoder = latentStageEncoder
        self._quantizationHead = quantizationHead
        self._latentHead = latentHead

    def syncCodebook(self):
        self._quantizer.syncCodebook()

    def reAssignCodebook(self, freq: torch.Tensor) -> torch.Tensor:
        return self._quantizer.reAssignCodebook(freq)

    def encode(self, x: torch.Tensor):
        # [h, w] -> [h/2, w/2]
        z = self._latentStageEncoder(x)
        code = self._quantizer.encode(self._quantizationHead(z))
        if self._latentHead is None:
            return None, code
        z = self._latentHead(z)
        #      ↓ residual,                         [n, m, h, w]
        return z - self._dequantizer.decode(code), code

    def forward(self, x: torch.Tensor):
        # [h, w] -> [h/2, w/2]
        z = self._latentStageEncoder(x)
        q, code, oneHot, logit = self._quantizer(self._quantizationHead(z))
        if self._latentHead is None:
            return q, None, code, oneHot, logit
        z = self._latentHead(z)
        #         ↓ residual
        return q, z - self._dequantizer(q), code, oneHot, logit

class _quantizerDecoder(nn.Module):
    """
    Default structure:
    ```plain
        q [H/2, W/2]            formerLevelRestored [H/2, W/2]
        | `dequantizaitonHead`  | `sideHead`
        ├-`add` ----------------╯
        xHat [H/2, W/2]
        | `restoreHead`
        nextLevelRestored [H, W]
    ```
    """

    def __init__(self, dequantizer: _multiCodebookDeQuantization, dequantizationHead: nn.Module, sideHead: Union[None, nn.Module], restoreHead: nn.Module):
        super().__init__()
        self._dequantizer =  dequantizer
        self._dequantizationHead =  dequantizationHead
        self._sideHead =  sideHead
        self._restoreHead =  restoreHead

    #                [n, m, h, w]
    def decode(self, code: torch.Tensor, formerLevel: Union[None, torch.Tensor]):
        q = self._dequantizationHead(self._dequantizer.decode(code))
        if self._sideHead is not None:
            xHat = q + self._sideHead(formerLevel)
        else:
            xHat = q
        return self._restoreHead(xHat)

    def forward(self, q: torch.Tensor, formerLevel: Union[None, torch.Tensor]):
        q = self._dequantizationHead(self._dequantizer(q))
        if self._sideHead is not None:
            xHat = q + self._sideHead(formerLevel)
        else:
            xHat = q
        return self._restoreHead(xHat)


class UMGMQuantizer(BaseQuantizer):
    _components = [
        "latentStageEncoder",
        "quantizationHead",
        "latentHead",
        "dequantizationHead",
        "sideHead",
        "restoreHead"
    ]
    def __init__(self, channel: int, m: int, k: Union[int, List[int]], components: Dict[str, Callable[[], nn.Module]]):
        if isinstance(k, int):
            k = [k]
        super().__init__(m, k)
        componentFns = [components[key] for key in self._components]
        latentStageEncoderFn, quantizationHeadFn, latentHeadFn, dequantizationHeadFn, sideHeadFn, restoreHeadFn = componentFns

        encoders = list()
        decoders = list()

        for i, ki in enumerate(k):
            latentStageEncoder = latentStageEncoderFn()
            quantizationHead = quantizationHeadFn()
            latentHead = latentHeadFn() if i < len(k) - 1 else None
            dequantizationHead = dequantizationHeadFn()
            sideHead = sideHeadFn() if i < len(k) - 1 else None
            restoreHead = restoreHeadFn()
            codebook = nn.Parameter(nn.init.normal_(torch.empty(m, ki, channel // m), std=math.sqrt(2 / (5 * channel / m))))
            quantizer = _multiCodebookQuantization(codebook)
            dequantizer = _multiCodebookDeQuantization(codebook)
            encoders.append(_quantizerEncoder(quantizer, dequantizer, latentStageEncoder, quantizationHead, latentHead))
            decoders.append(_quantizerDecoder(dequantizer, dequantizationHead, sideHead, restoreHead))

        self._encoders: nn.ModuleList[_quantizerEncoder] = nn.ModuleList(encoders)
        self._decoders: nn.ModuleList[_quantizerDecoder] = nn.ModuleList(decoders)

    def encode(self, x: torch.Tensor) -> List[torch.Tensor]:
        codes = list()
        for encoder in self._encoders:
            x, code = encoder.encode(x)
            #            [n, m, h, w]
            codes.append(code)
        # lv * [n, m, h, w]
        return codes

    def decode(self, codes: List[torch.Tensor]) -> Union[torch.Tensor, None]:
        formerLevel = None
        for decoder, code in zip(self._decoders[::-1], codes[::-1]):
            formerLevel = decoder.decode(code, formerLevel)
        return formerLevel

    def reAssignCodebook(self) -> torch.Tensor:
        freqs = self.Freq
        reassigned: List[torch.Tensor] = list()
        for encoder, freq in zip(self._encoders, freqs):
            # freq: [m, ki]
            reassigned.append(encoder.reAssignCodebook(freq))
        return torch.cat(reassigned).float().mean()

    def syncCodebook(self):
        dist.barrier()
        for encoder in self._encoders:
            encoder.syncCodebook()

    def forward(self, x: torch.Tensor):
        quantizeds = list()
        codes = list()
        oneHots = list()
        logits = list()
        for encoder in self._encoders:
            #          ↓ residual
            quantized, x, code, oneHot, logit = encoder(x)
            # [n, c, h, w]
            quantizeds.append(quantized)
            # [n, m, h, w]
            codes.append(code)
            # [n, m, h, w, k]
            oneHots.append(oneHot)
            # [n, m, h, w, k]
            logits.append(logit)
        formerLevel = None
        for decoder, quantized in zip(self._decoders[::-1], quantizeds[::-1]):
            # ↓ restored
            formerLevel = decoder(quantized, formerLevel)

        # update freq in entropy coder
        self._entropyCoder(oneHots)

        return formerLevel, codes, logits
